from dataclasses import dataclass
from typing import Union

from ..metadata import Area, Curve, ContinuousContract, SpecificContract
from .ohlc import OHLCList
from .timeseries import Timeseries
from .periodseries import Periodseries


@dataclass(frozen=True)
class SRMCOptions:
    """
    A data class with all factors for an SRMC calculations.
    """
    #: The fuel type. Set to either COAL or GAS.
    fuel_type: str = None
    #: (For coal) Conversion factor between API#2 tonnes and megawatthours.
    api2_tonne_to_mwh: float = None
    #: (For gas) Conversion from pence/therm to GBP/MWh.
    gas_therm_to_mwh: float = None
    #: The energy efficiency factor.
    efficiency: float = None
    #: The carbon emission factor.
    carbon_emissions: float = None
    #: (optional) When set, carbon tax is calculated according to tax rules
    #: in given area. Typically used for Great Britain's flat carbon tax.
    carbon_tax_area: Area = None

    def __str__(self):
        if self.fuel_type == 'GAS':
            return (
                f"<SRMCOptions: {self.fuel_type}, "
                f"gas_therm_to_mwh={self.gas_therm_to_mwh}, "
                f"efficiency={self.efficiency}, "
                f"carbon_emissions={self.carbon_emissions}, "
                f"carbon_tax_area={self.carbon_tax_area}"
                f">"
            )
        elif self.fuel_type == 'COAL':
            return (
                f"<SRMCOptions: {self.fuel_type}, "
                f"api2_tonne_to_mwh={self.api2_tonne_to_mwh}, "
                f"efficiency={self.efficiency}, "
                f"carbon_emissions={self.carbon_emissions}, "
                f"carbon_tax_area={self.carbon_tax_area}"
                f">"
            )
        else:
            raise ValueError("fuel_type must be either GAS or COAL")

    def __repr__(self):
        return self.__str__()


@dataclass(frozen=True)
class SRMC:
    """
    A data class for the API response from an SRMC calculation. It can
    either contain a list of OHLC objects, a time series, or a period-based
    series.
    """
    #: The curve used as input to this SRMC calculation
    curve: Curve = None
    #: (optional) The contract when the calculation was done for historical
    #: closing prices – either a front contract (continuous front contract),
    #: or a specific contract such as Year-2020, Year-2021, Q3-2020 etc.
    contract: Union[ContinuousContract, SpecificContract] = None
    #: The factors used in the SRMC calculation.
    options: SRMCOptions = None
    #: (optional) The OHLC data when the result of the calculation returns
    #: a list of OHLC objects.
    ohlc: OHLCList = None
    #: (optional) The time series when the result of the calculation returns
    #: a time series.
    timeseries: Timeseries = None
    #: (optional) The period-based series when the result of the calculation
    #: returns a period series.
    periodseries: Periodseries = None

    def has_ohlc(self):
        """
        Check whether this SRMC calculation did return OHLC objects.

        :return: True when this SRMC calculation did return OHLC objects,\
            otherwise False
        :rtype: bool
        """
        return self.ohlc is not None

    def has_timeseries(self):
        """
        Check whether this SRMC calculation did return a time series.

        :return: True when this SRMC calculation did return a time series,\
            otherwise False
        :rtype: bool
        """
        return self.timeseries is not None

    def has_periodseries(self):
        """
        Check whether this SRMC calculation did return a period-based series.

        :return: True when this SRMC calculation did return a period-based\
            series, otherwise False
        :rtype: bool
        """
        return self.periodseries is not None

    def __str__(self):
        parts = []
        if self.curve:
            parts.append(f"curve={self.curve}")
        if self.contract:
            parts.append(f"contract={self.contract}")
        if self.options:
            parts.append(f"options={self.options}")
        if self.has_ohlc():
            parts.append(f"ohlc={self.ohlc}")
        if self.has_timeseries():
            parts.append(f"timeseries={self.timeseries}")
        if self.has_periodseries():
            parts.append(f"periodseries={self.periodseries}")
        if not parts:
            return f"<SRMC: (no data)>"
        joined = ',\n '.join(parts)
        return f"<SRMC:\n {joined}\n>"

    def __repr__(self):
        return self.__str__()


from ..utils import Page, dict_to_str
from .base import BaseAPI

from ..parser.metadata import parse_curve, parse_place


class _MetadataAPI(BaseAPI):
    """
    Common metadata API operations.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cache = {}

    def curves(
            self,
            q=None,
            area=None,
            data_type=None,
            category=None,
            exact_category=None,
            frequency=None,
            commodity=None,
            has_place=None,
            page=1,
            page_size=50):
        """
        Search for curves. Supports paging.

        :param q: Freetext search, defaults to None
        :type q: str, optional
        :param area: Filter on an area, defaults to None
        :type area: Area, str, optional
        :param data_type: Filter on a DataType, defaults to None
        :type data_type: DataType, str, optional
        :param category: List of one or more categories that the curves\
            must have, defaults to None
        :type category: list[str], optional
        :param exact_category: An exact string of the category list for\
            a curve, defaults to None
        :type exact_category: str, optional
        :param frequency: Filter by frequency, defaults to None
        :type frequency: Frequency, str, optional
        :param commodity: Filter by commodity, defaults to None
        :type commodity: str, optional
        :param has_place: True – only curves for places, False – only curves\
            without places, defaults to None
        :type has_place: bool, optional
        :param page: Set the page, defaults to 1
        :type page: int, optional
        :param page_size: Set the page size, defaults to 50
        :type page_size: int, optional
        :return: A page of :py:class:`energyquantified.metadata.Curve`
        :rtype: :py:class:`energyquantified.utils.Page` (subclass of list)
        """
        # Parameters
        params = {}
        self._add_str(params, "q", q)
        self._add_area(params, "area", area)
        self._add_data_type(params, "data-type", data_type)
        self._add_str_list(params, "category", category)
        self._add_str(params, "exact-category", exact_category)
        self._add_str(params, "commodity", commodity)
        self._add_frequency(params, "frequency", frequency)
        self._add_bool(params, "has-place", has_place)
        self._add_int(params, "page", page)
        self._add_int(params, "page-size", page_size)
        # Load function
        def _load(page=None):
            # Override page if it is set
            if page:
                self._add_int(params, "page", page)
            # Check cache to see if we already have done this
            cache_key = dict_to_str(params, "curves")
            if self._cache.get(cache_key) is not None:
                return self._cache[cache_key]
            # Do the HTTP request, cache the result and return
            response = self._get("/metadata/curves/", params=params)
            items_gen = (parse_curve(c) for c in response.json())
            self._cache[cache_key] = (
                Page._response_to_page(items_gen, response, load_func=_load)
            )
            return self._cache[cache_key]
        # Invoke load
        return _load()


class MetadataAPI(_MetadataAPI):
    """
    Operations for curve search and place lookups. It also includes
    operations for listing attributes on
    :py:class:`energyquantified.metadata.Curve` objects.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def categories(self):
        """
        List all available categories.

        Results are cached on repeated calls.

        :return: A set of available categories (strings)
        :rtype: set
        """
        CACHE_KEY = "categories"
        if self._cache.get(CACHE_KEY):
            return self._cache[CACHE_KEY]
        response = self._get("/metadata/categories/")
        self._cache[CACHE_KEY] = set(response.json())
        return self._cache[CACHE_KEY]

    def exact_categories(self):
        """
        List available exact combinations of categories (meaning
        combinations that exists in curve names). You can use this response
        for filtering curves on the ``exact_category`` parameter in
        :py:meth:`MetadataAPI.curves`.

        Results are cached on repeated calls.

        :return: A set of category combinations (strings)
        :rtype: set
        """
        CACHE_KEY = "exact-categories"
        if self._cache.get(CACHE_KEY):
            return self._cache[CACHE_KEY]
        response = self._get("/metadata/exact-categories/")
        self._cache[CACHE_KEY] = set(response.json())
        return self._cache[CACHE_KEY]

    def places(
            self,
            q=None,
            area=None,
            fuel=None,
            kind=None,
            tree=False,
            page=1,
            page_size=50):
        """
        Search for places. Supports places.

        :param q: Freetext search, defaults to None
        :type q: str, optional
        :param area: Filter on an area, defaults to None
        :type area: Area, str, optional
        :param fuel: Filter by fuel type, defaults to None
        :type fuel: str, optional
        :param kind: Type of place (producer, consumer, river, etc.),\
            defaults to None
        :type kind: str, optional
        :param tree: Enable to return as a hierachy, defaults to False
        :type tree: bool, optional
        :param page: Set the page, defaults to 1
        :type page: int, optional
        :param page_size: Set the page size, defaults to 50
        :type page_size: int, optional
        :return: A page of :py:class:`energyquantified.metadata.Place`
        :rtype: :py:class:`energyquantified.utils.Page` (subclass of list)
        """
        # Parameters
        params = {}
        self._add_str(params, "q", q)
        self._add_area(params, "area", area)
        self._add_str_list(params, "fuel", fuel)
        self._add_str_list(params, "kind", kind)
        self._add_bool(params, "tree", tree)
        self._add_int(params, "page", page)
        self._add_int(params, "page-size", page_size)
        # Load function
        def _load(page=None):
            # Override page if it is set
            if page:
                self._add_int(params, "page", page)
            # Check cache to see if we already have done this
            cache_key = dict_to_str(params, "places")
            if self._cache.get(cache_key):
                return self._cache[cache_key]
            # Do the HTTP request, cache the result and return
            response = self._get("/metadata/places/", params=params)
            items_gen = (parse_place(p) for p in response.json())
            self._cache[cache_key] = (
                Page._response_to_page(items_gen, response, load_func=_load)
            )
            return self._cache[cache_key]
        # Invoke load
        return _load()


class RealtoMetadataAPI(_MetadataAPI):
    """
    The metadata API for Realto users. Has only the curves() method enabled.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from .area import Allocation, Area, Border
from .curve import Curve, CurveType, DataType
from .instance import Instance
from .place import Place, PlaceType
from .options import Aggregation, Filter
from .ohlc import (
    OHLCField,
    ContractPeriod,
    ContinuousContract,
    SpecificContract
)


__all__ = [
    # Area
    "Allocation",
    "Area",
    "Border",
    # Curve
    "Curve",
    "CurveType",
    "DataType",
    # Instance
    "Instance",
    # Places
    "Place",
    "PlaceType",
    # Options
    "Aggregation",
    "Filter",
    # OHLC
    "OHLCField",
    "ContractPeriod",
    "ContinuousContract",
    "SpecificContract"
]


from dateutil import parser

from ..data import Period, CapacityPeriod, Periodseries, PeriodseriesList
from ..exceptions import ParseException
from ..time import to_timezone
from .metadata import parse_curve, parse_instance, parse_resolution


def parse_periodseries_list(json):
    """
    Parse a JSON list-response from the API a list of Periodseries objects with
    Curve, Resolution and Instance.
    """
    if not isinstance(json, list):
        raise ParseException(
            f"Expected list of period-based series JSON objects, found: {type(json)}"
        )
    return PeriodseriesList(parse_periodseries(json_obj) for json_obj in json)


def parse_periodseries(json):
    """
    Parse a JSON response from the API into a Periodseries object with
    Curve, Resolution and Instance.
    """
    # Parse the Resolution
    resolution = json.get("resolution")
    if resolution:
        resolution = parse_resolution(resolution)
    # Parse the Curve
    curve = json.get("curve")
    if curve:
        curve = parse_curve(curve)
    # Parse the instance
    instance = json.get("instance")
    if instance:
        instance = parse_instance(instance, curve=curve)
    # Parse data
    data = _parse_data(json.get("data"), resolution=resolution)

    return Periodseries(
        curve=curve,
        resolution=resolution,
        instance=instance,
        data=data
    )


def _parse_data(json, resolution=None):
    """
    Parse the data[] part of the json response
    """
    # Empty time series
    if not json:
        return []
    # Look at the first element
    element = json[0]
    has_capacity = "capacity" in element
    if has_capacity:
        return _parse_value_capacity_data(json, resolution=resolution)
    else:
        return _parse_value_data(json, resolution=resolution)


def _parse_value_capacity_data(json, resolution=None):
    timezone = resolution.timezone

    data = []
    for element in json:
        begin = to_timezone(
            parser.isoparse(element.get("begin")),
            tz=timezone
        )
        end = to_timezone(
            parser.isoparse(element.get("end")),
            tz=timezone
        )
        value = element.get("v")
        installed = element.get("capacity")
        data.append(CapacityPeriod(begin, end, value, installed))

    return data


def _parse_value_data(json, resolution=None):
    timezone = resolution.timezone

    data = []
    for element in json:
        begin = to_timezone(
            parser.isoparse(element.get("begin")),
            tz=timezone
        )
        end = to_timezone(
            parser.isoparse(element.get("end")),
            tz=timezone
        )
        value = element.get("v")
        data.append(Period(begin, end, value))

    return data


_timeseries_class = None
def _get_timeseries_class():
    """
    Private utility function for lazy-loading the Timeseries class.

    :return: The Timeseries class
    :rtype: class
    """
    global _timeseries_class
    if not _timeseries_class:
        from energyquantified.data import Timeseries
        _timeseries_class = Timeseries
    return _timeseries_class

_ohlc_list_class = None
def _get_ohlc_list_class():
    """
    Private utility function for lazy-loading the OHLCList class.

    :return: The OHLCList class
    :rtype: class
    """
    global _ohlc_list_class
    if not _ohlc_list_class:
        from energyquantified.data import OHLCList
        _ohlc_list_class = OHLCList
    return _ohlc_list_class


_value_type_class = None
def _get_value_type_class():
    """
    Private utility function for lazy-loading the ValueType enum class.

    :return: The ValueType class
    :rtype: class
    """
    global _value_type_class
    if not _value_type_class:
        from energyquantified.data import ValueType
        _value_type_class = ValueType
    return _value_type_class


pd = None
_is_pandas_installed = None
def is_pandas_installed():
    """
    Private utility to both lazy-load and to check if pandas is installed.

    :return: True when pandas is available on the system, otherwise False
    :rtype: bool
    """
    global pd, _is_pandas_installed
    if _is_pandas_installed is None:
        try:
            import pandas as pd
            _is_pandas_installed = True
        except ImportError as e:
            _is_pandas_installed = False
    return _is_pandas_installed


def assert_pandas_installed():
    """
    Assert that pandas is installed.

    :raises ImportError: When pandas is not installed on the system
    """
    if not is_pandas_installed():
        if not is_pandas_installed():
            raise ImportError(
                "You must install the \"pandas\" data analysis library "
                "to use this functionality. Visit "
                "https://pandas.pydata.org/docs/ for more information."
            )


def timeseries_to_dataframe(timeseries, name=None, single_level_header=False):
    """
    Convert a time series to a ``pandas.DataFrame``.

    :param timeseries: The time series
    :type timeseries: energyquantified.data.Timeseries, required
    :param name: Set a custom name in the column header, Timeseries.curve.name \
        is used when this parameter is not set, defaults to None
    :type name: str, optional
    :param single_level_header: Set to True to use single-level header \
        in the DataFrame, defaults to False
    :type single_level_header: boolean, optional
    :return: A DataFrame
    :rtype: pandas.DataFrame
    :raises ImportError: When pandas is not installed on the system
    """
    # Checks
    assert_pandas_installed()
    assert isinstance(timeseries, _get_timeseries_class()), (
        "timeseries must be an instance of energyquantified.data.Timeseries"
    )
    assert isinstance(name, (type(None), str)), (
        f"parameter name must be string or None, was {type(name)}"
    )
    if name is None:
        name = timeseries.name
        include_instance = True
    else:
        include_instance = False
    # Conversion
    ValueType = _get_value_type_class()
    if timeseries.value_type() == ValueType.VALUE:
        # Convert a time series of (date, value)
        if single_level_header:
            return _timeseries_to_dataframe_value_single_header(
                timeseries,
                name,
                include_instance=include_instance
            )
        else:
            return _timeseries_to_dataframe_value(
                timeseries,
                name
            )
    if timeseries.value_type() == ValueType.SCENARIOS:
        # Convert a time series of (date, scenarios[])
        if single_level_header:
            return _timeseries_to_dataframe_scenarios_single_header(
                timeseries,
                name,
                include_instance=include_instance
            )
        else:
            return _timeseries_to_dataframe_scenarios(
                timeseries,
                name
            )
    if timeseries.value_type() == ValueType.MEAN_AND_SCENARIOS:
        # Convert a time series of (date, value, scenarios[])
        if single_level_header:
            return _timeseries_to_dataframe_mean_and_scenarios_single_header(
                timeseries,
                name,
                include_instance=include_instance
            )
        else:
            return _timeseries_to_dataframe_mean_and_scenarios(
                timeseries,
                name
            )
    # Unknown value type for time series
    raise ValueError(
        "Unknown ValueType: timeseries.value_type = "
        f"{timeseries.value_type()}"
    )


def _timeseries_to_dataframe_value(timeseries, name):
    """
    Private utility function for converting a time series of single values
    to a pandas dataframe.

    The DataFrame will have a three-level column header.

    :param timeseries: A time series object
    :type timeseries: Timeseries
    :param name: The time series name
    :type name: str
    :return: A pandas DataFrame
    :rtype: pandas.DataFrame
    """
    # Column headers
    columns = [
        [name],
        [timeseries.instance_or_contract_dataframe_column_header()],
        ['']
    ]
    # Convert a time series of (date, value)
    df = pd.DataFrame.from_records(
        ((v.value,) for v in timeseries),
        columns=columns,
        index=[v.date for v in timeseries],
    )
    df.index.name = 'date'
    return df


def _timeseries_to_dataframe_value_single_header(timeseries, name,
        include_instance=True):
    """
    Private utility function for converting a time series of single values
    to a pandas dataframe.

    The DateFrame will have a single-level column header.

    :param timeseries: A time series object
    :type timeseries: Timeseries
    :param name: The time series name
    :type name: str
    :param include_instance: Include the instance in the header?
    :type include_instance: bool
    :return: A pandas DataFrame
    :rtype: pandas.DataFrame
    """
    # Column header
    if include_instance:
        instance = timeseries.instance_or_contract_dataframe_column_header()
        columns = [
            [f"{name} {instance}".strip()]
        ]
    else:
        columns = [
            [name]
        ]
    # Convert a time series of (date, value)
    df = pd.DataFrame.from_records(
        ((v.value,) for v in timeseries),
        columns=columns,
        index=[v.date for v in timeseries],
    )
    df.index.name = 'date'
    return df


def _timeseries_to_dataframe_scenarios(timeseries, name):
    """
    Private utility function for converting a time series of scenario values
    to a pandas dataframe.

    The DataFrame will have a three-level column header.

    :param timeseries: A time series object
    :type timeseries: Timeseries
    :param name: The time series name
    :type name: str
    :return: A pandas DataFrame
    :rtype: pandas.DataFrame
    """
    width = timeseries.total_values_per_item()
    # Column headers
    columns = [
        [name] * width,
        [timeseries.instance_or_contract_dataframe_column_header()] * width,
        timeseries.scenario_names
    ]
    # Convert a time series of (date, scenarios[])
    df = pd.DataFrame.from_records(
        (v.scenarios for v in timeseries.data),
        columns=columns,
        index=[v.date for v in timeseries],
    )
    df.index.name = 'date'
    return df


def _timeseries_to_dataframe_scenarios_single_header(timeseries, name,
        include_instance=True):
    """
    Private utility function for converting a time series of scenario values
    to a pandas dataframe.

    The DateFrame will have a single-level column header.

    :param timeseries: A time series object
    :type timeseries: Timeseries
    :param name: The time series name
    :type name: str
    :param include_instance: Include the instance in the header?
    :type include_instance: bool
    :return: A pandas DataFrame
    :rtype: pandas.DataFrame
    """
    width = timeseries.total_values_per_item()
    # Column header
    if include_instance:
        instance = timeseries.instance_or_contract_dataframe_column_header()
        columns = [
            [
                f"{name} {instance} {scenario}".strip()
                for scenario in timeseries.scenario_names
            ],
        ]
    else:
        columns = [
            [
                f"{name} {scenario}".strip()
                for scenario in timeseries.scenario_names
            ],
        ]
    # Convert a time series of (date, scenarios[])
    df = pd.DataFrame.from_records(
        (v.scenarios for v in timeseries.data),
        columns=columns,
        index=[v.date for v in timeseries],
    )
    df.index.name = 'date'
    return df


def _timeseries_to_dataframe_mean_and_scenarios(timeseries, name):
    """
    Private utility function for converting a time series of a mean value
    and scenarios to a pandas dataframe.

    The DataFrame will have a three-level column header.

    :param timeseries: A time series object
    :type timeseries: Timeseries
    :param name: The time series name
    :type name: str
    :return: A pandas DataFrame
    :rtype: pandas.DataFrame
    """
    width = timeseries.total_values_per_item()
    # Column headers
    columns = [
        [name] * width,
        [timeseries.instance_or_contract_dataframe_column_header()] * width,
        [''] + timeseries.scenario_names
    ]
    # Convert a time series of (date, scenarios[])
    df = pd.DataFrame.from_records(
        ((v.value, *v.scenarios) for v in timeseries.data),
        columns=columns,
        index=[v.date for v in timeseries],
    )
    df.index.name = 'date'
    return df


def _timeseries_to_dataframe_mean_and_scenarios_single_header(timeseries, name,
        include_instance=True):
    """
    Private utility function for converting a time series of a mean value
    and scenarios to a pandas dataframe.

    The DateFrame will have a single-level column header.

    :param timeseries: A time series object
    :type timeseries: Timeseries
    :param name: The time series name
    :type name: str
    :param include_instance: Include the instance in the header?
    :type include_instance: bool
    :return: A pandas DataFrame
    :rtype: pandas.DataFrame
    """
    width = timeseries.total_values_per_item()
    # Column headers
    scenario_names = [''] + timeseries.scenario_names
    if include_instance:
        instance = timeseries.instance_or_contract_dataframe_column_header()
        columns = [
            [
                f"{name} {instance} {scenario}".strip()
                for scenario in scenario_names
            ],
        ]
    else:
        columns = [
            [f"{name} {scenario}".strip() for scenario in scenario_names],
        ]
    # Convert a time series of (date, scenarios[])
    df = pd.DataFrame.from_records(
        ((v.value, *v.scenarios) for v in timeseries.data),
        columns=columns,
        index=[v.date for v in timeseries],
    )
    df.index.name = 'date'
    return df


def timeseries_list_to_dataframe(timeseries_list, single_level_header=False):
    """
    Convert a list of time series to a ``pandas.DataFrame``.

    :param timeseries_list: [description]
    :type timeseries_list: [type]
    :param single_level_header: Set to True to use single-level header \
        in the DataFrame, defaults to False
    :type single_level_header: boolean, optional
    :return: A pandas DataFrame
    :rtype: pandas.DataFrame
    """
    # Checks
    assert_pandas_installed()
    assert timeseries_list, "timeseries list is empty"
    for index, timeseries in enumerate(timeseries_list):
        assert isinstance(timeseries, _get_timeseries_class()), (
            f"timeseries_list[{index}] must be an instance of "
            f"energyquantified.data.Timeseries, but was: {type(timeseries)}"
        )
    # Merge into one data frame
    return pd.concat(
        [
            ts.to_dataframe(single_level_header=single_level_header)
            for ts in timeseries_list
        ],
        axis=1
    )


def ohlc_list_to_dataframe(ohlc_list):
    """
    Convert an :py:class:`energyquantified.data.OHLCList` to a
    ``pandas.DataFrame``.

    :param ohlc_list: A list of OHLC objects
    :type ohlc_list: OHLCList
    :return: A DataFrame
    :rtype: pandas.DataFrame
    :raises ImportError: When pandas is not installed on the system
    """
    # Checks
    assert_pandas_installed()
    assert isinstance(ohlc_list, _get_ohlc_list_class()), (
        "ohlc_list must be an instance of energyquantified.data.OHLCList"
    )
    # Conversion
    return pd.DataFrame.from_records((
        {
            "traded": ohlc.product.traded,
            "period": ohlc.product.period.tag,
            "front": ohlc.product.front,
            "delivery": ohlc.product.delivery,
            "open": ohlc.open,
            "high": ohlc.high,
            "low": ohlc.low,
            "close": ohlc.close,
            "settlement": ohlc.settlement,
            "volume": ohlc.volume,
            "open_interest": ohlc.open_interest,
        } for ohlc in ohlc_list
    ))


import enum


class Place:
    """
    A place is a general concept of a physical location, such as a powerplant,
    a weather station, a position on a river etc.
    """

    def __init__(self, kind, key, name, unit=None, fuels=None, area=None,
                 areas=None, location=None, children=None, curves=None):
        #: The place type. See :py:class:`PlaceType`.
        self.kind = kind
        #: The identifier
        self.key = key
        #: The name of the place
        self.name = name
        #: The unit name (if it is a powerplant unit)
        self.unit = unit
        #: The fuel types (if it is a powerplant unit)
        self.fuels = fuels or []
        #: The area in which this place lies, see :py:class:`Area`. This
        #: field is deprecated and will be removed in future releases.
        self.area = area
        #: The list areas in which this place lies (mulitple when it's on a
        #: border), see :py:class:`Area`.
        self.areas = areas or []
        #: The geolocation of this place: ``(latitude, longitude)``
        self.location = location or None
        #: A list of children (typically used for a powerplants with sub-units)
        self.children = children or []
        #: A list of curves with data for this place. See :py:class:`Curve`.
        self.curves = curves or []

    @property
    def latitude(self):
        """
        The latitude of this place.
        """
        return self.location[0]

    @property
    def longitude(self):
        """
        The longitude of this place.
        """
        return self.location[1]

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        parts = []
        parts.append(f"<Place: key=\"{self.key}\", name=\"{self.name}\"")
        if self.unit:
            parts.append(f", unit=\"{self.unit}\"")
        if self.kind:
            parts.append(f", kind={self.kind}")
        if self.fuels:
            parts.append(f", fuels={self.fuels}")
        if self.location:
            parts.append(f", location={self.location}")
        parts.append(">")
        return "".join(parts)


_placetypes_lookup = {}

class PlaceType(enum.Enum):
    """
    Enumerator of place types. Used to describe type type of a
    :py:class:`Place`.
    """
    #: A city
    CITY = ("city",)
    #: A power consumer, such as a factory
    CONSUMER = ("consumer",)
    #: A power producer (power plant)
    PRODUCER = ("producer",)
    #: A river location
    RIVER = ("river",)
    #: A weather station
    WEATHERSTATION = ("weatherstation",)
    #: Gas storage
    GAS_STORAGE = ("gas-storage",)
    #: Gas LNG terminal
    GAS_LNG_TERMINAL = ("gas-lng-terminal",)
    #: Gas interconnector
    GAS_INTERCONNECTOR = ("gas-interconnector",)
    #: Unspecified
    OTHER = ("other",)

    def __init__(self, tag):
        self.tag = tag
        _placetypes_lookup[tag.lower()] = self

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return self.name

    @staticmethod
    def is_valid_tag(tag):
        """
        Check whether a place type tag exists or not.

        :param tag: A place type tag
        :type tag: str
        :return: True if it exists, otherwise False
        :rtype: bool
        """
        return tag.lower() in _placetypes_lookup

    @staticmethod
    def by_tag(tag):
        """
        Look up place type by tag.

        :param tag: A place type tag
        :type tag: str
        :return: The place type for the given tag
        :rtype: PlaceType
        """
        return _placetypes_lookup[tag.lower()]



from collections import namedtuple
from dataclasses import dataclass
from datetime import date
import enum

from .base import Series
from ..metadata import ContractPeriod, OHLCField
from ..time import Frequency
from ..utils.pandas import ohlc_list_to_dataframe



@dataclass(frozen=True)
class Product:
    """
    An energy product description with a trading date and the
    contract in detail.
    """
    traded: date
    period: ContractPeriod
    front: int
    delivery: date

    def __str__(self):
        return (
            f"<Product: traded={self.traded.isoformat()}, "
            f"period={self.period.name}, "
            f"front={self.front}, "
            f"delivery={self.delivery.isoformat()}>"
        )

    def __repr__(self):
        return self.__str__()


@dataclass(frozen=True)
class OHLC:
    """
    A summary for a trading day on a contract.
    """
    product: Product
    open: float
    high: float
    low: float
    close: float
    settlement: float
    volume: float
    open_interest: float

    def get_field(self, field):
        """
        Get an OHLC field from this object.

        :param field: A OHLC field
        :type field: OHLCField | str
        :raises ValueError: When the `field` parameter isn't a valid OHLC field
        :return: The value on the given OHLC field or None if it isn't set
        :rtype: float
        """
        if isinstance(field, str) and OHLCField.is_valid_tag(field):
            return getattr(self, field, default=None)
        elif isinstance(field, OHLCField):
            return getattr(self, field.tag, default=None)
        else:
            raise ValueError(f'Unknown field: {field}')

    def __str__(self):
        return (
            f"<OHLC: {self.product}, "
            f"open={self.open or ''}, "
            f"high={self.high or ''}, "
            f"low={self.low or ''}, "
            f"close={self.close or ''}, "
            f"settlement={self.settlement or ''}, "
            f"volume={self.volume or ''}, "
            f"open_interest={self.open_interest or ''}>"
        )

    def __repr__(self):
        return self.__str__()


class OHLCList(list):
    """
    A collection of OHLC data. Can contain all sorts of contracts
    (yearly, monthly, weekly etc.) for a specific market.
    """

    def __init__(self, elements, curve=None, contract=None):
        super().__init__(elements)
        # --- Public members ---
        #: The curve holding these OHLC objects
        self.curve = curve

    def append(self, value):
        raise NotImplementedError("OHLCList does not support append")

    def extend(self, iterable):
        raise NotImplementedError("OHLCList does not support extend")

    def insert(self, index, value):
        raise NotImplementedError("OHLCList does not support insert")

    def remove(self, value):
        raise NotImplementedError("OHLCList does not support remove")

    def pop(self, index=-1):
        raise NotImplementedError("OHLCList does not support pop")

    def clear(self):
        raise NotImplementedError("OHLCList does not support clear")

    def __setitem__(self, key, value):
        raise NotImplementedError("OHLCList does not support __setitem__")

    def __delitem__(self, key):
        raise NotImplementedError("OHLCList does not support __delitem__")

    def __add__(self, rhs):
        raise NotImplementedError("OHLCList does not support __add__")

    def __iadd__(self, rhs):
        raise NotImplementedError("OHLCList does not support __iadd__")

    def __mul__(self, rhs):
        raise NotImplementedError("OHLCList does not support multiply")

    def __rmul__(self, rhs):
        raise NotImplementedError("OHLCList does not support multiply")

    def __imul__(self, rhs):
        raise NotImplementedError("OHLCList does not support multiply")

    def __str__(self):
        if self.curve:
            return f"<OHLCList: curve=\"{self.curve}\", items={super().__str__()}>"
        else:
            return f"<OHLCList: items={super().__str__()}>"

    def to_df(self):
        """
        Convert this OHLCList to a ``pandas.DataFrame`` with a column for
        open, high, low, close, settlement, volume and open interest.

        :return: A DataFrame
        :rtype: pandas.DataFrame
        :raises ImportError: When pandas is not installed on the system
        """
        return self.to_dataframe()

    def to_dataframe(self):
        """
        Convert this OHLCList to a ``pandas.DataFrame`` with a column for
        open, high, low, close, settlement, volume and open interest.

        :return: A DataFrame
        :rtype: pandas.DataFrame
        :raises ImportError: When pandas is not installed on the system
        """
        return ohlc_list_to_dataframe(self)



from .instances import InstancesAPI
from .metadata import MetadataAPI, RealtoMetadataAPI
from .timeseries import TimeseriesAPI
from .periods import PeriodsAPI
from .period_instances import PeriodInstancesAPI
from .ohlc import OhlcAPI
from .srmc import SrmcAPI

__all__ = [
    "InstancesAPI",
    "MetadataAPI",
    "RealtoMetadataAPI",
    "TimeseriesAPI",
    "PeriodsAPI",
    "PeriodInstancesAPI",
    "OhlcAPI",
    "SrmcAPI",
]


from .base import BaseAPI

from ..metadata import CurveType
from ..parser.timeseries import parse_timeseries


# Tuple of supported values for Curve.curve_type in the time series API
CURVE_TYPES = (CurveType.TIMESERIES, CurveType.SCENARIO_TIMESERIES)

class TimeseriesAPI(BaseAPI):
    """
    Time series API operations. Access these operations via an
    instance of the :py:class:`energyquantified.EnergyQuantified` class:

       >>> eq = EnergyQuantified(api_key="aaaa-bbbb-cccc-dddd")
       >>> eq.timeseries.load(curve, begin, end)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def load(
            self,
            curve,
            begin=None,
            end=None,
            frequency=None,
            aggregation=None,
            hour_filter=None):
        """
        Load time series data for a :py:class:`energyquantified.metadata.Curve`.

        This operation works for curves with
        ``curve_type = TIMESERIES | SCENARIO_TIMESERIES`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param begin: The begin date-time
        :type begin: date, datetime, str, required
        :param end: The end date-time
        :type end: date, datetime, str, required
        :param frequency: Set the preferred frequency for aggregations, defaults to None
        :type frequency: Frequency, optional
        :param aggregation: The aggregation method (i.e. AVERAGE, MIN, MAX),\
            has no effect unless *frequency* is provided, defaults to AVERAGE
        :type aggregation: Aggregation, optional
        :param hour_filter: Filters on hours to include (i.e. BASE, PEAK),\
            has no effect unless *frequency* is provided, defaults to BASE
        :type hour_filter: Filter, optional
        :return: A time series
        :rtype: :py:class:`energyquantified.data.Timeseries`
        """
        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        url = f"/timeseries/{safe_curve}/"
        # Parameters
        params = {}
        self._add_datetime(params, "begin", begin, required=True)
        self._add_datetime(params, "end", end, required=True)
        self._add_frequency(params, "frequency", frequency)
        if "frequency" in params:
            self._add_aggregation(params, "aggregation", aggregation)
            self._add_filter(params, "hour-filter", hour_filter)
        # HTTP request
        response = self._get(url, params=params)
        return parse_timeseries(response.json())



from .base import BaseAPI

from ..exceptions import ValidationError
from ..metadata import Curve, CurveType, OHLCField
from ..parser.srmc import parse_srmc_response


# Tuple of supported values for Curve.curve_type in the time series API
CURVE_TYPES = (CurveType.OHLC,)

class SrmcAPI(BaseAPI):
    """
    Operations for SRMC calculations in the API. Access these operations via
    an instance of the :py:class:`energyquantified.EnergyQuantified` class:

       >>> eq = EnergyQuantified(api_key="aaaa-bbbb-cccc-dddd")
       >>> eq.srmc.load_front(curve, begin, end, period='month', front=1)
    """

    @staticmethod
    def _check_curve_category(curve):
        # Check that the curve has either 'gas' or 'coal' in its name
        if isinstance(curve, Curve):
            name = curve.name.lower()
            if not 'gas' in name and not 'coal' in name:
                raise ValidationError(
                    reason="Provide a coal or gas curve",
                    parameter="curve"
                )
        elif isinstance(curve, str):
            name = curve.lower()
            if not 'gas' in name and not 'coal' in name:
                raise ValidationError(
                    reason="Provide a coal or gas curve",
                    parameter="curve"
                )

    def load_front(
            self,
            curve,
            begin=None,
            end=None,
            period=None,
            front=None,
            gas_therm_to_mwh=None,
            api2_tonne_to_mwh=None,
            carbon_emissions=None,
            efficiency=None,
            carbon_tax_area=None):
        """
        Calculate historical short-run margincal costs (SRMC) for a
        continuous front contract.

        SRMC is calculated from a **coal** or **gas** curve of your choosing.
        It uses the daily reference rates from the European Central Bank (ECB)
        for currency conversions. The EUA price in the calculation is the
        settlement price from ICE.

        Some countries, such as Great Britain, has an additional flat tax on
        carbon emissions. Specify the ``carbon_tax_area`` parameter to apply
        tax rules for a specific country.

        This operation works for **coal** or **gas** curves with
        ``curve_type = OHLC`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param begin: The begin date
        :type begin: date, str, required
        :param end: The end date
        :type end: date, str, required
        :param period: Filter on contract period (day, week, month etc.),\
            defaults to None
        :type period: ContractPeriod, str, required
        :param front: The front contract (1=front, 2=second front, etc.)
        :type front: int, required
        :param gas_therm_to_mwh: Conversion from pence/therm to GBP/MWh\
            (still in higher-heating value). Defaults to 0.029307.
        :type gas_therm_to_mwh: float, optional
        :param api2_tonne_to_mwh: Conversion from API2 coal to MWh.\
            Defaults to 6.978.
        :type api2_tonne_to_mwh: float, optional
        :param carbon_emissions: The carbon content as tCO2/MWh. This value\
            varies between coal and gas. For coal, the default factor is\
            0.34056. For gas, the default factor is 0.202.
        :type carbon_emissions: float, optional
        :param efficiency: The energy efficiency. For coal, the default\
            factor is 0.42. For gas, the default factor is 0.59.
        :type efficiency: float, optional
        :param carbon_tax_area: Set an area to apply tax rules for.
        :type carbon_tax_area: Area, str, optional
        :return: An SRMC object with a list of OHLC objects
        :rtype: :py:class:`energyquantified.data.SRMC`
        """
        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        self._check_curve_category(curve)
        url = f"/srmc/{safe_curve}/"
        # Parameters
        params = {}
        self._add_date(params, "begin", begin, required=True)
        self._add_date(params, "end", end, required=True)
        self._add_contract_period(params, "period", period, required=True)
        self._add_int(params, "front", front, min=1, required=True)
        self._add_number(params, "gas-therm-to-mwh", gas_therm_to_mwh)
        self._add_number(params, "api2-tonne-to-mwh", api2_tonne_to_mwh)
        self._add_number(params, "carbon-emissions", carbon_emissions, min=0.01, max=1.0)
        self._add_number(params, "efficiency", efficiency, min=0.01, max=1.0)
        self._add_area(params, "carbon-tax-area", carbon_tax_area)
        # HTTP request
        response = self._get(url, params=params)
        return parse_srmc_response(response.json())

    def load_delivery(
            self,
            curve,
            begin=None,
            end=None,
            period=None,
            delivery=None,
            gas_therm_to_mwh=None,
            api2_tonne_to_mwh=None,
            carbon_emissions=None,
            efficiency=None,
            carbon_tax_area=None):
        """
        Calculate historical short-run margincal costs (SRMC) for a specific
        contract, such as the year 2021, the month Jan 2021, etc.

        SRMC is calculated from a **coal** or **gas** curve of your choosing.
        It uses the daily reference rates from the European Central Bank (ECB)
        for currency conversions. The EUA price in the calculation is the
        settlement price from ICE.

        Some countries, such as Great Britain, has an additional flat tax on
        carbon emissions. Specify the ``carbon_tax_area`` parameter to apply
        tax rules for a specific country.

        This operation works for **coal** or **gas** curves with
        ``curve_type = OHLC`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param begin: The begin date
        :type begin: date, str, required
        :param end: The end date
        :type end: date, str, required
        :param period: Filter on contract period (day, week, month etc.),\
            defaults to None
        :type period: ContractPeriod, str, required
        :param delivery: Filter on delivery date, requires parameter ``period``\
            to be set; cannot be used together with ``front``, defaults to None
        :type delivery: date, str, required
        :param gas_therm_to_mwh: Conversion from pence/therm to GBP/MWh\
            (still in higher-heating value). Defaults to 0.029307.
        :type gas_therm_to_mwh: float, optional
        :param api2_tonne_to_mwh: Conversion from API2 coal to MWh.\
            Defaults to 6.978.
        :type api2_tonne_to_mwh: float, optional
        :param carbon_emissions: The carbon content as tCO2/MWh. This value\
            varies between coal and gas. For coal, the default factor is\
            0.34056. For gas, the default factor is 0.202.
        :type carbon_emissions: float, optional
        :param efficiency: The energy efficiency. For coal, the default\
            factor is 0.42. For gas, the default factor is 0.59.
        :type efficiency: float, optional
        :param carbon_tax_area: Set an area to apply tax rules for.
        :type carbon_tax_area: Area, str, optional
        :return: An SRMC object with a list of OHLC objects
        :rtype: :py:class:`energyquantified.data.SRMC`
        """
        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        self._check_curve_category(curve)
        url = f"/srmc/{safe_curve}/"
        # Parameters
        params = {}
        self._add_date(params, "begin", begin, required=True)
        self._add_date(params, "end", end, required=True)
        self._add_contract_period(params, "period", period, required=True)
        self._add_date(params, "delivery", delivery, required=True)
        self._add_number(params, "gas-therm-to-mwh", gas_therm_to_mwh)
        self._add_number(params, "api2-tonne-to-mwh", api2_tonne_to_mwh)
        self._add_number(params, "carbon-emissions", carbon_emissions, min=0.01, max=1.0)
        self._add_number(params, "efficiency", efficiency, min=0.01, max=1.0)
        self._add_area(params, "carbon-tax-area", carbon_tax_area)
        # HTTP request
        response = self._get(url, params=params)
        return parse_srmc_response(response.json())

    def load_front_as_timeseries(
            self,
            curve,
            begin=None,
            end=None,
            frequency=None,
            period=None,
            front=None,
            fill=None,
            gas_therm_to_mwh=None,
            api2_tonne_to_mwh=None,
            carbon_emissions=None,
            efficiency=None,
            carbon_tax_area=None):
        """
        Calculate historical short-run margincal costs (SRMC) for a continuous
        front contract and convert the result to a daily time series.

        SRMC is calculated from a **coal** or **gas** curve of your choosing.
        It uses the daily reference rates from the European Central Bank (ECB)
        for currency conversions. The EUA price in the calculation is the
        settlement price from ICE.

        Some countries, such as Great Britain, has an additional flat tax on
        carbon emissions. Specify the ``carbon_tax_area`` parameter to apply
        tax rules for a specific country.

        This operation works for **coal** or **gas** curves with
        ``curve_type = OHLC`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param begin: The begin date
        :type begin: date, str, required
        :param end: The end date
        :type end: date, str, required
        :param period: Filter on contract period (day, week, month etc.),\
            defaults to None
        :type period: ContractPeriod, str, required
        :param front: The front contract (1=front, 2=second front, etc.)
        :type front: int, required
        :param fill: How to handle days without trades. Allowed values are:\
            ``no-fill`` do nothing, ``fill-holes`` fill in holes with data\
            from previous trading day, ``forward-fill`` fill in all blanks\
            with data from the previous trading day (also into the future).\
            Defaults to ``no-fill``.
        :type fill: str, optional
        :param gas_therm_to_mwh: Conversion from pence/therm to GBP/MWh\
            (still in higher-heating value). Defaults to 0.029307.
        :type gas_therm_to_mwh: float, optional
        :param api2_tonne_to_mwh: Conversion from API2 coal to MWh.\
            Defaults to 6.978.
        :type api2_tonne_to_mwh: float, optional
        :param carbon_emissions: The carbon content as tCO2/MWh. This value\
            varies between coal and gas. For coal, the default factor is\
            0.34056. For gas, the default factor is 0.202.
        :type carbon_emissions: float, optional
        :param efficiency: The energy efficiency. For coal, the default\
            factor is 0.42. For gas, the default factor is 0.59.
        :type efficiency: float, optional
        :param carbon_tax_area: Set an area to apply tax rules for.
        :type carbon_tax_area: Area, str, optional
        :return: An SRMC object with a ``timeseries`` object
        :rtype: :py:class:`energyquantified.data.SRMC`
        """
        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        self._check_curve_category(curve)
        url = f"/srmc/{safe_curve}/timeseries/"
        # Parameters
        params = {}
        self._add_date(params, "begin", begin, required=True)
        self._add_date(params, "end", end, required=True)
        self._add_contract_period(params, "period", period, required=True)
        self._add_int(params, "front", front, min=1, required=True)
        self._add_fill(params, "fill", fill)
        self._add_number(params, "gas-therm-to-mwh", gas_therm_to_mwh)
        self._add_number(params, "api2-tonne-to-mwh", api2_tonne_to_mwh)
        self._add_number(params, "carbon-emissions", carbon_emissions, min=0.01, max=1.0)
        self._add_number(params, "efficiency", efficiency, min=0.01, max=1.0)
        self._add_area(params, "carbon-tax-area", carbon_tax_area)
        # HTTP request
        response = self._get(url, params=params)
        return parse_srmc_response(response.json())

    def load_delivery_as_timeseries(
            self,
            curve,
            begin=None,
            end=None,
            period=None,
            delivery=None,
            fill=None,
            gas_therm_to_mwh=None,
            api2_tonne_to_mwh=None,
            carbon_emissions=None,
            efficiency=None,
            carbon_tax_area=None):
        """
        Calculate historical short-run margincal costs (SRMC) for a specific
        contract and convert it to a daily time series.

        A specific contract could be the year 2021, the month Jan 2021, etc.

        SRMC is calculated from a **coal** or **gas** curve of your choosing.
        It uses the daily reference rates from the European Central Bank (ECB)
        for currency conversions. The EUA price in the calculation is the
        settlement price from ICE.

        Some countries, such as Great Britain, has an additional flat tax on
        carbon emissions. Specify the ``carbon_tax_area`` parameter to apply
        tax rules for a specific country.

        This operation works for **coal** or **gas** curves with
        ``curve_type = OHLC`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param begin: The begin date
        :type begin: date, str, required
        :param end: The end date
        :type end: date, str, required
        :param period: Filter on contract period (day, week, month etc.),\
            defaults to None
        :type period: ContractPeriod, str, required
        :param delivery: Filter on delivery date, requires parameter ``period``\
            to be set; cannot be used together with ``front``, defaults to None
        :type delivery: date, str, required
        :param fill: How to handle days without trades. Allowed values are:\
            ``no-fill`` do nothing, ``fill-holes`` fill in holes with data\
            from previous trading day, ``forward-fill`` fill in all blanks\
            with data from the previous trading day (also into the future).\
            Defaults to ``no-fill``.
        :type fill: str, optional
        :param gas_therm_to_mwh: Conversion from pence/therm to GBP/MWh\
            (still in higher-heating value). Defaults to 0.029307.
        :type gas_therm_to_mwh: float, optional
        :param api2_tonne_to_mwh: Conversion from API2 coal to MWh.\
            Defaults to 6.978.
        :type api2_tonne_to_mwh: float, optional
        :param carbon_emissions: The carbon content as tCO2/MWh. This value\
            varies between coal and gas. For coal, the default factor is\
            0.34056. For gas, the default factor is 0.202.
        :type carbon_emissions: float, optional
        :param efficiency: The energy efficiency. For coal, the default\
            factor is 0.42. For gas, the default factor is 0.59.
        :type efficiency: float, optional
        :param carbon_tax_area: Set an area to apply tax rules for.
        :type carbon_tax_area: Area, str, optional
        :return: An SRMC object with a ``timeseries`` object
        :rtype: :py:class:`energyquantified.data.SRMC`
        """
        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        self._check_curve_category(curve)
        url = f"/srmc/{safe_curve}/timeseries/"
        # Parameters
        params = {}
        self._add_date(params, "begin", begin, required=True)
        self._add_date(params, "end", end, required=True)
        self._add_contract_period(params, "period", period, required=True)
        self._add_date(params, "delivery", delivery, required=True)
        self._add_fill(params, "fill", fill)
        self._add_number(params, "gas-therm-to-mwh", gas_therm_to_mwh)
        self._add_number(params, "api2-tonne-to-mwh", api2_tonne_to_mwh)
        self._add_number(params, "carbon-emissions", carbon_emissions, min=0.01, max=1.0)
        self._add_number(params, "efficiency", efficiency, min=0.01, max=1.0)
        self._add_area(params, "carbon-tax-area", carbon_tax_area)
        # HTTP request
        response = self._get(url, params=params)
        return parse_srmc_response(response.json())

    def latest(
            self,
            curve,
            date=None,
            gas_therm_to_mwh=None,
            api2_tonne_to_mwh=None,
            carbon_emissions=None,
            efficiency=None,
            carbon_tax_area=None):
        """
        Calculate short-run margincal costs (SRMC) for all settlement prices
        from a trading day. Defaults to using OHLC data from the latest
        available trading day, hence ``latest()``.

        If ``date`` is given, this method will try to fetch OHLC data for
        that trading day. When there is no data for the given day, OHLC data
        will be loaded for the closest trading day earlier in time with data.

        SRMC is calculated from a **coal** or **gas** curve of your choosing.
        It uses the daily reference rates from the European Central Bank (ECB)
        for currency conversions. The EUA price in the calculation is the
        settlement price from ICE.

        Some countries, such as Great Britain, has an additional flat tax on
        carbon emissions. Specify the ``carbon_tax_area`` parameter to apply
        tax rules for a specific country.

        This operation works for **coal** or **gas** curves with
        ``curve_type = OHLC`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param date: The trading date, defaults to today
        :type date: date, str, required
        :param gas_therm_to_mwh: Conversion from pence/therm to GBP/MWh\
            (still in higher-heating value). Defaults to 0.029307.
        :type gas_therm_to_mwh: float, optional
        :param api2_tonne_to_mwh: Conversion from API2 coal to MWh.\
            Defaults to 6.978.
        :type api2_tonne_to_mwh: float, optional
        :param carbon_emissions: The carbon content as tCO2/MWh. This value\
            varies between coal and gas. For coal, the default factor is\
            0.34056. For gas, the default factor is 0.202.
        :type carbon_emissions: float, optional
        :param efficiency: The energy efficiency. For coal, the default\
            factor is 0.42. For gas, the default factor is 0.59.
        :type efficiency: float, optional
        :param carbon_tax_area: Set an area to apply tax rules for.
        :type carbon_tax_area: Area, str, optional
        :return: An SRMC object with a list of OHLC objects
        :rtype: :py:class:`energyquantified.data.SRMC`
        """
        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        self._check_curve_category(curve)
        url = f"/srmc/{safe_curve}/latest/"
        # Parameters
        params = {}
        self._add_date(params, "date", date)
        self._add_number(params, "gas-therm-to-mwh", gas_therm_to_mwh)
        self._add_number(params, "api2-tonne-to-mwh", api2_tonne_to_mwh)
        self._add_number(params, "carbon-emissions", carbon_emissions, min=0.01, max=1.0)
        self._add_number(params, "efficiency", efficiency, min=0.01, max=1.0)
        self._add_area(params, "carbon-tax-area", carbon_tax_area)
        # HTTP request
        response = self._get(url, params=params)
        return parse_srmc_response(response.json())

    def latest_as_periods(
            self,
            curve,
            date=None,
            gas_therm_to_mwh=None,
            api2_tonne_to_mwh=None,
            carbon_emissions=None,
            efficiency=None,
            carbon_tax_area=None):
        """
        Calculate short-run margincal costs (SRMC) for all settlement prices
        from a trading day, sort them, and merge/convert them to a continuous
        series.

        Defaults to using OHLC data from the latest available trading day,
        hence ``latest`` in the method name.

        If ``date`` is given, this method will try to fetch OHLC data for
        that trading day. When there is no data for the given day, OHLC data
        will be loaded for the closest trading day earlier in time with data.

        SRMC is calculated from a **coal** or **gas** curve of your choosing.
        It uses the daily reference rates from the European Central Bank (ECB)
        for currency conversions. The EUA price in the calculation is the
        settlement price from ICE.

        Some countries, such as Great Britain, has an additional flat tax on
        carbon emissions. Specify the ``carbon_tax_area`` parameter to apply
        tax rules for a specific country.

        This operation works for **coal** or **gas** curves with
        ``curve_type = OHLC`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param date: The trading date, defaults to today
        :type date: date, str, required
        :param gas_therm_to_mwh: Conversion from pence/therm to GBP/MWh\
            (still in higher-heating value). Defaults to 0.029307.
        :type gas_therm_to_mwh: float, optional
        :param api2_tonne_to_mwh: Conversion from API2 coal to MWh.\
            Defaults to 6.978.
        :type api2_tonne_to_mwh: float, optional
        :param carbon_emissions: The carbon content as tCO2/MWh. This value\
            varies between coal and gas. For coal, the default factor is\
            0.34056. For gas, the default factor is 0.202.
        :type carbon_emissions: float, optional
        :param efficiency: The energy efficiency. For coal, the default\
            factor is 0.42. For gas, the default factor is 0.59.
        :type efficiency: float, optional
        :param carbon_tax_area: Set an area to apply tax rules for.
        :type carbon_tax_area: Area, str, optional
        :return: An SRMC object with a period-based series
        :rtype: :py:class:`energyquantified.data.SRMC`
        """
        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        self._check_curve_category(curve)
        url = f"/srmc/{safe_curve}/latest/periods/"
        # Parameters
        params = {}
        self._add_date(params, "date", date)
        self._add_number(params, "gas-therm-to-mwh", gas_therm_to_mwh)
        self._add_number(params, "api2-tonne-to-mwh", api2_tonne_to_mwh)
        self._add_number(params, "carbon-emissions", carbon_emissions, min=0.01, max=1.0)
        self._add_number(params, "efficiency", efficiency, min=0.01, max=1.0)
        self._add_area(params, "carbon-tax-area", carbon_tax_area)
        # HTTP request
        response = self._get(url, params=params)
        return parse_srmc_response(response.json())





class Series:
    """
    Base class for Timeseries and Periodseries.
    """

    def __init__(
            self,
            curve=None,
            name=None,
            resolution=None,
            instance=None,
            contract=None):
        # --- Members ---
        #: The curve
        self.curve = curve
        self._name = None
        #: The resolution
        self.resolution = resolution
        #: The instance (if any)
        self.instance = instance
        #: The contract for OHLC operations resulting in a time series
        self.contract = contract
        #: The data
        self.data = []

    @property
    def name(self):
        """
        Return the curve name, if a :class:`energyquantified.metadata.Curve`
        is associated with this series.

        :return: The ``curve.name`` if it exists, otherwise ``None``
        :rtype: str, NoneType
        """
        if self._name:
            return self._name
        if self.curve:
            return self.curve.name
        return None

    def set_name(self, name):
        """
        Set a custom name (defaults to using ``curve.name``).

        :param name: An user-defined name
        :type name: str
        """
        assert name is None or isinstance(name, str),\
            "name must be None or a string"
        self._name = name

    def instance_or_contract_dataframe_column_header(self):
        """
        Get the instance or contract for this time series, in a format
        fitting for a ``pandas.DataFrame`` column header.

        :return: An instance or contract column header for a data frame
        :rtype: str, NoneType
        """
        if self.instance:
            return self.instance.as_dataframe_column_header()
        if self.contract:
            return self.contract.as_dataframe_column_header()
        return ''

    def has_data(self):
        """
        Check if this series has any data.

        :return: True if there is at least one data point, otherwise False
        :rtype: bool
        """
        return bool(self.data)

    def begin(self):
        """
        Get the begin date-time in the timeseries (inclusive).

        :return: The begin date
        :rtype: datetime
        """
        # This method must be implemented in subclasses
        raise NotImplementedError("Method begin() is not implemented")

    def end(self):
        """
        Get the end date-time in the timeseries (exclusive).

        :return: The end date
        :rtype: datetime
        """
        # This method must be implemented in subclasses
        raise NotImplementedError("Method end() is not implemented")

    def __iter__(self):
        """
        Create a generator of all values in the time series.

        :return: A generator for all data points in the time series
        :rtype: generator
        """
        return (item for item in self.data or [])



from dateutil import parser

from ..data import Product, OHLC, OHLCList
from ..exceptions import ParseException
from ..metadata import ContractPeriod
from .metadata import parse_curve, parse_contract


def parse_ohlc_response(json):
    # Parse curve
    curve = json.get("curve")
    if curve:
        curve = parse_curve(curve)
    # Parse OHLC objects
    data = parse_ohlc_list(json.get("data"))

    return OHLCList(data, curve=curve)


def parse_ohlc_list(json):
    # Parse data[], but keep track of data[index] to create a better exception
    data = []
    ohlc = None
    for index, item in enumerate(json):
        try:
            ohlc = parse_ohlc(item)
        except Exception as e:
            raise ParseException(
                f"Failed to parse data[{index}] in OHLC JSON response: {ohlc}"
            ) from e
        data.append(ohlc)
    return data


def parse_ohlc(json):
    product = parse_product(json.get("product"))
    return OHLC(
        product=product,
        open=json.get("open"),
        high=json.get("high"),
        low=json.get("low"),
        close=json.get("close"),
        settlement=json.get("settlement"),
        volume=json.get("volume"),
        open_interest=json.get("open_interest")
    )


def parse_product(json):
    traded = parser.isoparse(json.get("traded_at")).date()
    period = ContractPeriod.by_tag(json.get("period"))
    front = json.get("front")
    delivery = parser.isoparse(json.get("delivery")).date()
    return Product(
        traded=traded,
        period=period,
        front=front,
        delivery=delivery
    )



from datetime import datetime, date, timedelta

from dateutil.relativedelta import relativedelta

from .utils import is_tz_aware


def shift(
    datetime_obj,
    years=0,
    months=0,
    weeks=0,
    days=0,
    hours=0,
    minutes=0,
    seconds=0,
    millis=0,
):
    """
    Shift a date by a certain amount of time, using a timedelta
    under the hood, while normalizing the tzinfo.
    """
    # Make the delta
    delta = relativedelta(
        years=years,
        months=months,
        weeks=weeks,
        days=days,
        hours=hours,
        minutes=minutes,
        seconds=seconds,
        microseconds=millis * 1000,
    )
    if years != 0 or months != 0 or weeks != 0 or days != 0:
        # Do arithmetic in without time-zone info, then localize when done
        tz, dt_naive = datetime_obj.tzinfo, datetime_obj.replace(tzinfo=None)
        dt_naive = dt_naive + delta
        return tz.localize(dt_naive)
    else:
        # Add delta and normalize datetime if tzinfo is attached
        if is_tz_aware(datetime_obj):
            return datetime_obj.tzinfo.normalize(datetime_obj + delta)
        return datetime_obj + delta


def replace(datetime_obj, **kwargs):
    """
    Replaces fields of a datetime while normalizing the tzinfo. All
    kwargs will be passed on to datetime.replace.
    """
    if is_tz_aware(datetime_obj):
        return datetime_obj.tzinfo.normalize(datetime_obj.replace(**kwargs))
    return datetime_obj.replace(**kwargs)


def truncate_second(datetime_obj):
    """
    Truncates and returns new a datetime to seconds.

    Keeps the timezone of the datetime. Works for timezone naive objects, too.
    """
    d = datetime_obj.replace(microsecond=0)
    if is_tz_aware(datetime_obj):
        return datetime_obj.tzinfo.normalize(d)
    return d


def truncate_minute(datetime_obj):
    """
    Truncates and returns new a datetime to minutes.

    Keeps the timezone of the datetime. Works for timezone naive objects, too.
    """
    d = datetime_obj.replace(microsecond=0, second=0)
    if is_tz_aware(datetime_obj):
        return datetime_obj.tzinfo.normalize(d)
    return d


def truncate_5min(datetime_obj):
    """
    Truncates and returns new a datetime to 5 minutes (:00, :05, :10, ...).

    Keeps the timezone of the datetime. Works for timezone naive objects, too.
    """
    m = (datetime_obj.minute // 5) * 5
    d = datetime_obj.replace(microsecond=0, second=0, minute=m)
    if is_tz_aware(datetime_obj):
        return datetime_obj.tzinfo.normalize(d)
    return d


def truncate_10min(datetime_obj):
    """
    Truncates and returns new a datetime to 10 minutes (:00, :10, :20, ...).

    Keeps the timezone of the datetime. Works for timezone naive objects, too.
    """
    m = (datetime_obj.minute // 10) * 10
    d = datetime_obj.replace(microsecond=0, second=0, minute=m)
    if is_tz_aware(datetime_obj):
        return datetime_obj.tzinfo.normalize(d)
    return d


def truncate_15min(datetime_obj):
    """
    Truncates and returns new a datetime to 15 minutes (:00, :15, :30, :45).

    Keeps the timezone of the datetime. Works for timezone naive objects, too.
    """
    m = (datetime_obj.minute // 15) * 15
    d = datetime_obj.replace(microsecond=0, second=0, minute=m)
    if is_tz_aware(datetime_obj):
        return datetime_obj.tzinfo.normalize(d)
    return d


def truncate_30min(datetime_obj):
    """
    Truncates and returns new a datetime to 30 minutes (:00, :30).

    Keeps the timezone of the datetime. Works for timezone naive objects, too.
    """
    m = (datetime_obj.minute // 30) * 30
    d = datetime_obj.replace(microsecond=0, second=0, minute=m)
    if is_tz_aware(datetime_obj):
        return datetime_obj.tzinfo.normalize(d)
    return d


def truncate_hour(datetime_obj):
    """
    Truncates and returns new a datetime to hours.

    Keeps the timezone of the datetime. Works for timezone naive objects, too.
    """
    d = datetime_obj.replace(microsecond=0, second=0, minute=0)
    if is_tz_aware(datetime_obj):
        return datetime_obj.tzinfo.normalize(d)
    return d


def truncate_date(datetime_obj):
    """
    Truncates and returns new a datetime to date, or returns the date
    if it is a date.

    Keeps the timezone of the datetime. Works for timezone naive objects, too.
    """
    if isinstance(datetime_obj, datetime):
        # Datetime
        d = datetime_obj.replace(microsecond=0, second=0, minute=0, hour=0)
        if is_tz_aware(datetime_obj):
            return datetime_obj.tzinfo.normalize(d)
        return d
    else:
        # Date
        return datetime_obj


def truncate_week(datetime_obj):
    """
    Truncates and returns new a datetime to date, or returns the date
    if it is a date.

    Keeps the timezone of the datetime. Works for timezone naive objects, too.
    """
    days_from_monday = datetime_obj.isoweekday() - 1
    # Date
    if isinstance(datetime_obj, datetime):
        # Datetime
        d = datetime_obj.replace(microsecond=0, second=0, minute=0, hour=0)
        d = d - timedelta(days=days_from_monday)
        if is_tz_aware(datetime_obj):
            return datetime_obj.tzinfo.normalize(d)
        return d
    else:
        # Date
        return datetime_obj - timedelta(days=days_from_monday)


def truncate_month(datetime_obj):
    """
    Truncates and returns new a datetime to first of month, or returns the date
    if it is a date.

    Keeps the timezone of the datetime. Works for timezone naive objects, too.
    """
    if isinstance(datetime_obj, datetime):
        # Datetime
        d = datetime_obj.replace(microsecond=0, second=0, minute=0, hour=0, day=1)
        if is_tz_aware(datetime_obj):
            return datetime_obj.tzinfo.normalize(d)
        return d
    else:
        # Date
        return datetime_obj.replace(day=1)


def truncate_quarter(datetime_obj):
    """
    Truncates and returns new a datetime to the first day of a quarter, or
    returns the date if it is a date.

    Keeps the timezone of the datetime. Works for timezone naive objects, too.
    """
    month = datetime_obj.month
    new_month = month - ((month - 1) % 3)
    if isinstance(datetime_obj, datetime):
        # Datetime
        d = datetime_obj.replace(
            microsecond=0, second=0, minute=0, hour=0, day=1, month=new_month
        )
        if is_tz_aware(datetime_obj):
            return datetime_obj.tzinfo.normalize(d)
        return d
    else:
        # Date
        return datetime_obj.replace(day=1, month=new_month)


def truncate_season(datetime_obj):
    """
    Truncates and returns new a datetime to the first day of a season, or
    returns the date if it is a date.

    Keeps the timezone of the datetime. Works for timezone naive objects, too.
    """
    month = datetime_obj.month
    year = datetime_obj.year
    new_month = month - ((month - 1) % 3)
    # Check the various months
    if new_month == 1:
        new_month = 10
        year = year - 1
    elif new_month == 10:
        pass
    else:
        new_month = 4
    # Handle date and datetime objects differently
    if isinstance(datetime_obj, datetime):
        # Datetime
        d = datetime_obj.replace(
            microsecond=0, second=0, minute=0, hour=0, day=1, month=new_month, year=year
        )
        if is_tz_aware(datetime_obj):
            return datetime_obj.tzinfo.normalize(d)
        return d
    else:
        # Date
        return datetime_obj.replace(day=1, month=new_month, year=year)


def truncate_year(datetime_obj):
    """
    Truncates and returns new a datetime for the first day of the year, or
    returns the date if it is a date.

    Keeps the timezone of the datetime. Works for timezone naive objects, too.
    """
    if isinstance(datetime_obj, datetime):
        # Datetime
        d = datetime_obj.replace(
            microsecond=0, second=0, minute=0, hour=0, day=1, month=1
        )
        if is_tz_aware(datetime_obj):
            return datetime_obj.tzinfo.normalize(d)
        return d
    else:
        # Date
        return datetime_obj.replace(day=1, month=1)



import urllib
from datetime import date, datetime, time
import numbers

from requests import exceptions

from ..metadata import (
    Area,
    Curve,
    DataType,
    Aggregation,
    Filter,
    ContractPeriod,
    OHLCField,
)
from ..time import Frequency

from ..exceptions import (
    ForbiddenError,
    HTTPError,
    InternalServerError,
    NotFoundError,
    UnauthorizedError,
    ValidationError,
)


def _urlencode(text, safe=''):
    """
    Replace special characters in string using the %xx escape. Letters, digits,
    and the characters '_.-~' are never quoted. This function is intended for
    quoting the path section of a URL.

    The optional safe parameter specifies additional ASCII characters that
    should *not* be quoted — its default value is ''. If you want to leave i.e.
    '/' unquoted, you must set safe = '/'.

    :param text: A string to be quoted
    :type text: str, bytes
    :param safe: A string of safe characters, defaults to ''
    :type safe: str, optional
    :return: A quoted string
    :rtype: str, bytes
    """
    return urllib.parse.quote(text, safe=safe)


class BaseAPI:

    def __init__(self, client):
        self._client = client
        self._session = client._session

    def _get(self, url, params=None, ignore403=False, ignore404=False):
        # Perform HTTP request and check for any IO errors
        try:
            response = self._session.get(url, params=params)
        except exceptions.RequestException as e:
            raise HTTPError(e)

        # Check response for common errors
        if response.status_code == 400:
            # Parameters validation error
            json = response.json()
            raise ValidationError(
                reason=json.get("message"),
                parameter=json.get("parameter")
            )
        if response.status_code == 401:
            # Authentication error
            json = response.json()
            raise UnauthorizedError(reason=json.get("message"))
        if response.status_code == 403 and not ignore403:
            # Forbidden error
            json = response.json()
            raise ForbiddenError(reason=json.get("message"))
        if response.status_code == 404 and not ignore404:
            # Not found
            json = response.json()
            raise NotFoundError(reason=json.get("message"))
        if 500 <= response.status_code < 600:
            # Some internal server error
            raise InternalServerError()

        # Most likely OK at this point
        return response

    def _post(self, url, params, body):
        raise NotImplementedError

    @staticmethod
    def _urlencode_curve_name(curve, curve_types=None):
        if curve is None:
            raise ValidationError(
                reason="Provide a Curve or a curve name",
                parameter="curve"
            )
        curve_name = None
        if isinstance(curve, str):
            curve_name = curve
        elif isinstance(curve, Curve):
            if curve_types and curve.curve_type not in curve_types:
                raise ValidationError(
                    reason=(
                        f"Curve.curve_type must be any of {curve_types}, "
                        f"but was {curve.curve_type}"
                    ),
                    parameter="curve"
                )
            curve_name = curve.name
        else:
            raise ValidationError(
                reason="Provide either a string or a Curve instance",
                parameter="curve"
            )
        return _urlencode(curve_name)

    @staticmethod
    def _urlencode_datetime(datetime_obj, name):
        if datetime_obj is None:
            raise ValidationError(
                reason="Provide a date or datetime",
                parameter=name
            )
        date_string = None
        # Check for different types. IMPORTANT:
        # The isinstance(datetime) check must come before isinstance(date),
        # as a datetime is also an instance of date...
        if isinstance(datetime_obj, str):
            date_string = datetime_obj
        elif isinstance(datetime_obj, datetime):
            date_string = datetime_obj.isoformat(sep=" ")
        elif isinstance(datetime_obj, date):
            date_string = datetime_obj.isoformat()
        else:
            raise ValidationError(reason="Provide a datetime", parameter=name)
        return _urlencode(date_string)

    @staticmethod
    def _urlencode_string(text, name, transform_func=None):
        if text is None or not isinstance(text, str):
            raise ValidationError(
                reason="Provide a string",
                parameter=name
            )
        if transform_func:
            text = transform_func(text)
        return _urlencode(text)

    @staticmethod
    def _urlencode_ohlc_field(ohlc_field, name):
        if ohlc_field is None:
            raise ValidationError(
                reason="Provide an OHLCField",
                parameter=name
            )
        if isinstance(ohlc_field, str):
            field_string = ohlc_field
        elif isinstance(ohlc_field, OHLCField):
            field_string = ohlc_field.tag
        else:
            raise ValidationError(reason="Provide an OHLCField", parameter=name)
        return _urlencode(field_string)

    @staticmethod
    def _add_datetime(params, name, var, required=False):
        if var is None and not required:
            return
        if isinstance(var, str):
            params[name] = var
        elif isinstance(var, (date, datetime)):
            params[name] = var
        else:
            raise ValidationError(reason="Provide a datetime", parameter=name)

    @staticmethod
    def _add_date(params, name, var, required=False):
        if var is None and not required:
            return
        if isinstance(var, str):
            params[name] = var
        elif isinstance(var, date):
            params[name] = var
        else:
            raise ValidationError(reason="Provide a date", parameter=name)

    @staticmethod
    def _add_time(params, name, var, required=False):
        if var is None and not required:
            return
        if isinstance(var, str):
            params[name] = var
        elif isinstance(var, time):
            params[name] = var.isoformat(timespec='milliseconds')
        else:
            raise ValidationError(reason="Provide a time", parameter=name)

    @staticmethod
    def _add_str(params, name, var, allowed_values=None, required=False):
        if var is None and not required:
            return
        if isinstance(var, str):
            if allowed_values and not var in allowed_values:
                raise ValidationError(
                    reason=f"Allowed values are {allowed_values}",
                    parameter=name
                )
            params[name] = var
        else:
            raise ValidationError(reason="Provide a string", parameter=name)

    @staticmethod
    def _add_str_list(params, name, var, required=False):
        if (var is None or var == []) and not required:
            return
        if isinstance(var, str):
            params[name] = [var]
        elif isinstance(var, (list, tuple)):
            for element in var:
                if not isinstance(element, str):
                    raise ValidationError(
                        reason=(
                            f"All element in list must be "
                            f"strings, but found: {type(element)}"
                        ),
                        parameter=name
                    )
            params[name] = list(var)
        else:
            raise ValidationError(
                reason="Provide a string or a list of strings",
                parameter=name
            )

    @staticmethod
    def _add_int(params, name, var, required=False, min=None, max=None):
        if var is None and not required:
            return
        if isinstance(var, int):
            if min is not None and var < min:
                raise ValidationError(
                    reason=f"Must be higher than {min}, was {var}",
                    parameter=name
                )
            if max is not None and var > max:
                raise ValidationError(
                    reason=f"Must be lower than {max}, was {var}",
                    parameter=name
                )
            params[name] = var
        else:
            raise ValidationError(reason="Provide an integer", parameter=name)

    @staticmethod
    def _add_number(params, name, var, required=False, min=None, max=None):
        if var is None and not required:
            return
        if isinstance(var, numbers.Number):
            if min is not None and var < min:
                raise ValidationError(
                    reason=f"Must be higher than {min}, was {var}",
                    parameter=name
                )
            if max is not None and var > max:
                raise ValidationError(
                    reason=f"Must be lower than {max}, was {var}",
                    parameter=name
                )
            params[name] = var
        else:
            raise ValidationError(reason="Provide a number", parameter=name)

    @staticmethod
    def _add_bool(params, name, var, required=False):
        if var is None and not required:
            return
        if isinstance(var, bool):
            params[name] = 'true' if var else 'false'
        else:
            raise ValidationError(
                reason="Provide a boolean (True or False)",
                parameter=name
            )

    @staticmethod
    def _add_area(params, name, var, required=False):
        if var is None and not required:
            return
        if isinstance(var, str):
            if not Area.is_valid_tag(var):
                raise ValidationError(
                    reason=f"Not a valid Area: '{var}'",
                    parameter=name
                )
            params[name] = var
        elif isinstance(var, Area):
            params[name] = var.tag
        else:
            raise ValidationError(
                reason=f"Not a valid Area: '{var}'",
                parameter=name
            )

    @staticmethod
    def _add_data_type(params, name, var, required=False):
        if var is None and not required:
            return
        if isinstance(var, str):
            if not DataType.is_valid_tag(var):
                raise ValidationError(
                    reason=f"Not a valid DataType: '{var}'",
                    parameter=name
                )
            params[name] = var
        elif isinstance(var, DataType):
            params[name] = var.tag
        else:
            raise ValidationError(
                reason=f"Not a valid DataType: '{var}'",
                parameter=name
            )

    @staticmethod
    def _add_frequency(params, name, var, required=False):
        if var is None and not required:
            return
        if isinstance(var, str):
            if not Frequency.is_valid_tag(var):
                raise ValidationError(
                    reason=f"Not a valid Frequency: '{var}'",
                    parameter=name
                )
            params[name] = var
        elif isinstance(var, Frequency):
            params[name] = var.tag
        else:
            raise ValidationError(
                reason=f"Not a valid Frequency: '{var}'",
                parameter=name
            )

    @staticmethod
    def _add_aggregation(params, name, var, required=False):
        if var is None and not required:
            return
        if isinstance(var, str):
            if not Aggregation.is_valid_tag(var):
                raise ValidationError(
                    reason=f"Not a valid Aggregation: '{var}'",
                    parameter=name
                )
            params[name] = var
        elif isinstance(var, Aggregation):
            params[name] = var.tag
        else:
            raise ValidationError(
                reason=f"Not a valid Aggregation: '{var}'",
                parameter=name
            )

    @staticmethod
    def _add_filter(params, name, var, required=False):
        if var is None and not required:
            return
        if isinstance(var, str):
            if not Filter.is_valid_tag(var):
                raise ValidationError(
                    reason=f"Not a valid Filter: '{var}'",
                    parameter=name
                )
            params[name] = var
        elif isinstance(var, Filter):
            params[name] = var.tag
        else:
            raise ValidationError(
                reason=f"Not a valid Filter: '{var}'",
                parameter=name
            )

    @staticmethod
    def _add_contract_period(params, name, var, required=False):
        if var is None and not required:
            return
        if isinstance(var, str):
            if not ContractPeriod.is_valid_tag(var):
                raise ValidationError(
                    reason=f"Not a valid ContractPeriod: '{var}'",
                    parameter=name
                )
            params[name] = var
        elif isinstance(var, ContractPeriod):
            params[name] = var.tag
        else:
            raise ValidationError(
                reason=f"Not a valid ContractPeriod: '{var}'",
                parameter=name
            )

    @staticmethod
    def _add_fill(params, name, var, required=False):
        if var is None and not required:
            return
        if isinstance(var, str):
            var = var.lower()
            if var in ('no-fill', 'fill-holes', 'forward-fill'):
                params[name] = var
                return
        raise ValidationError(
            reason=(
                f"Not a valid fill '{var}'. Allowed values "
                f"are: 'no-fill', 'fill-holes', 'forward-fill'"
            ),
            parameter=name
        )



import time
from datetime import date, datetime

from .timezone import DEFAULT_TZ, UTC


def now(tz=DEFAULT_TZ):
    """
    Get the current datetime.

    :param tz: The preferred time-zone, defaults to DEFAULT_TZ
    :type tz: TzInfo (or similar pytz time-zone)
    :return: A time-zone aware datetime set to now
    :rtype: datetime
    """
    return datetime.now(tz=tz)

def today(tz=DEFAULT_TZ):
    """
    Get a datetime set to midnight today.

    :param tz: The preferred time-zone, defaults to DEFAULT_TZ
    :type tz: TzInfo (or similar pytz time-zone)
    :return: A time-zone aware datetime set to midnight this morning
    :rtype: datetime
    """

    dt = now(tz=tz)
    dt = dt.replace(microsecond=0, second=0, minute=0, hour=0)
    return dt.tzinfo.normalize(dt)

def get_date(year=None, month=1, day=1):
    """
    Get a date. Only year needs to be specified. The remaining fields
    defaults to the lowest value.

    :param year: The year
    :type year: int, required
    :param month: The month (1-12), defaults to 1
    :type month: int, optional
    :param day: The day-of-month (1-31), defaults to 1
    :type day: int, optional
    :return: A date
    :rtype: date
    """
    assert year and 1900 <= year <= 2100, "Must specify year (1900-2100)"
    return date(year, month, day)

def get_datetime(year=None, month=1, day=1, hour=0, minute=0, second=0,
        millis=0, tz=DEFAULT_TZ):
    """
    Get a datetime. Only year needs to be specified. The remaining fields
    defaults to the lowest value. Timezone defaults to `DEFAULT_TZ`.

    :param year: The year
    :type year: int, required
    :param month: The month (1-12), defaults to 1
    :type month: int, optional
    :param day: The day-of-month (1-31), defaults to 1
    :type day: int, optional
    :param hour: The hour-of-day (0-23), defaults to 0
    :type hour: int, optional
    :param minute: The minute (0-59), defaults to 0
    :type minute: int, optional
    :param second: The second (0-59), defaults to 0
    :type second: int, optional
    :param millis: The milliseconds (0-999), defaults to 0
    :type millis: int, optional
    :param tz: The time-zone, defaults to DEFAULT_TZ
    :type tz: TzInfo (or similar pytz time-zone)
    :return: A time-zone aware datetime
    :rtype: datetime
    """

    assert year and 1900 <= year <= 2100, "Must specify year (1900-2100)"
    dt = datetime(year, month, day, hour, minute, second, millis * 1000)
    return tz.localize(dt)


def to_millis(datetime_obj):
    """
    Convert a datetime object to milliseconds since epoch.
    """
    assert is_tz_aware(datetime_obj), "Must be timezone-aware"
    # First convert to UTC
    utc_dt = datetime_obj.astimezone(UTC)
    # Get seconds since epoch and fraction of a second in millis
    seconds = int(time.mktime(utc_dt.timetuple()) * 1000)
    millis_fraction = utc_dt.microsecond // 1000
    # Tada!
    return seconds + millis_fraction


def from_millis(millis, tz=DEFAULT_TZ):
    """
    Convert milliseconds since epoch to a datetime in given timezone.

    Timezone defaults to `DEFAULT_TZ`.
    """
    dt = datetime.fromtimestamp(millis / 1000.0, UTC)
    return to_timezone(dt, tz=tz)


def to_timezone(datetime_obj, tz=DEFAULT_TZ):
    """
    Make a datetime object timezone-aware in given timezone. If it already
    has a timezone, convert the instant to provided timezone.
    """
    if is_tz_aware(datetime_obj):
        return datetime_obj.astimezone(tz)
    return tz.localize(datetime_obj)


def is_tz_aware(datetime_obj):
    """
    Check whether a datetime_obj is timezone-aware or not.
    """
    tz = datetime_obj.tzinfo
    return tz is not None and tz.utcoffset(datetime_obj) is not None


def is_tz_naive(datetime_obj):
    """
    Check whether a datetime_obj is timezone-naive.
    """
    return not is_tz_aware(datetime_obj)


def is_in_timezone(datetime_obj, tz):
    """
    Check whether datetime_obj is in given timezone.
    """
    return datetime_obj.tzinfo == tz



from collections import namedtuple
from datetime import datetime
import numbers
import sys

from ..time import Resolution, Frequency
from .base import Series
from .timeseries import TimeseriesList, Timeseries, Value


class Period(namedtuple("Period", ("begin", "end", "value"))):
    """
    A period for a period-based series.

    Implemented as a namedtuple of (begin, end, value).

    .. py:attribute:: begin

        The begin date-time

    .. py:attribute:: end

        The end date-time

    .. py:attribute:: value

        The numeric value
    """

    def validate(self):
        """
        Check if this period is in a valid format.

        :raises ValueError: When this period tuple is invalid
        """
        if not isinstance(self.begin, datetime):
            raise ValueError("Period.begin is not a datetime")
        if not isinstance(self.end, datetime):
            raise ValueError("Period.end is not a datetime")
        if not (self.value is None or isinstance(self.value, (numbers.Number))):
            raise ValueError("Period.value must either be a number or None")

    def print(self, file=sys.stdout):
        """
        Print this period-based series value.
        """
        d0 = self.begin.isoformat(sep=' ')
        d1 = self.end.isoformat(sep=' ')
        v = self.value
        print(f"{d0}–{d1}\t{v:13.2f}", file=file)

    def to_dict(self):
        """
        Convert this period value to a dict in the same short
        format as the API response.

        :return: A dict of this period in the same format as the API response
        :rtype: dict
        """
        return {
            "begin": self.begin.isoformat(sep=' '),
            "end": self.end.isoformat(sep=' '),
            "v": self.value
        }

    def __str__(self):
        d0 = self.begin.isoformat(sep=' ')
        d1 = self.end.isoformat(sep=' ')
        v = self.value
        return f"<Period: begin={d0}, end={d1}, value={v}>"

    def __repr__(self):
        return str(self)

    def is_active(self, date):
        return self.begin <= date < self.end

    def is_empty_or_invalid(self):
        return self.begin >= self.end

    def is_date_before(self, date):
        return date < self.begin

    def is_date_after(self, date):
        return date >= self.end

    def is_interval_before(self, begin, end):
        return end <= self.begin

    def is_interval_after(self, begin, end):
        return begin >= self.end

    def is_overlayed(self, begin, end):
        return end > self.begin and begin < self.end

    def is_covered(self, begin, end):
        return self.begin <= begin and self.end >= end

    def get_duration_seconds(self, begin, end):
        d0 = max(self.begin, begin)
        d1 = min(self.end, end)
        return (d1 - d0).total_seconds()


class CapacityPeriod(namedtuple("Period", ("begin", "end", "value", "installed"))):
    """
    A period for a period-based series. Includes the installed capacity for
    the period (which may differ from the currently available capacity given in
    the *value* attribute).

    Implemented as a namedtuple of (begin, end, value).

    .. py:attribute:: begin

        The begin date-time

    .. py:attribute:: end

        The end date-time

    .. py:attribute:: value

        The numeric value

    .. py:attribute:: installed

        The total installed capacity
    """

    def validate(self):
        """
        Check if this period is in a valid format.

        :raises ValueError: When this period tuple is invalid
        """
        if not isinstance(self.begin, datetime):
            raise ValueError("Period.begin is not a datetime")
        if not isinstance(self.end, datetime):
            raise ValueError("Period.end is not a datetime")
        if not (self.value is None or isinstance(self.value, (numbers.Number))):
            raise ValueError("Period.value must either be a number or None")
        if not (self.installed is None or isinstance(self.installed, (numbers.Number))):
            raise ValueError("Period.installed must either be a number or None")

    def print(self, file=sys.stdout):
        """
        Print this period-based series value.
        """
        d0 = self.begin.isoformat(sep=' ')
        d1 = self.end.isoformat(sep=' ')
        v = self.value
        c = self.installed
        print(f"{d0}–{d1}\t{v:13.2f}\t{c:13.2f}", file=file)

    def to_dict(self):
        """
        Convert this period value to a dict in the same short
        format as the API response.

        :return: A dict of this period in the same format as the API response
        :rtype: dict
        """
        return {
            "begin": self.begin.isoformat(sep=' '),
            "end": self.end.isoformat(sep=' '),
            "v": self.value,
            "c": self.installed,
        }

    def __str__(self):
        d0 = self.begin.isoformat(sep=' ')
        d1 = self.end.isoformat(sep=' ')
        v = self.value
        c = self.installed
        return f"<Period: begin={d0}, end={d1}, value={v}, installed={c}>"

    def __repr__(self):
        return str(self)

    def is_active(self, date):
        return self.begin <= date < self.end

    def is_empty_or_invalid(self):
        return self.begin >= self.end

    def is_date_before(self, date):
        return date < self.begin

    def is_date_after(self, date):
        return date >= self.end

    def is_interval_before(self, begin, end):
        return end <= self.begin

    def is_interval_after(self, begin, end):
        return begin >= self.end

    def is_overlayed(self, begin, end):
        return end > self.begin and begin < self.end

    def is_covered(self, begin, end):
        return self.begin <= begin and self.end >= end

    def get_duration_seconds(self, begin, end):
        d0 = max(self.begin, begin)
        d1 = min(self.end, end)
        return (d1 - d0).total_seconds()


class Periodseries(Series):
    """
    A period-based series with metadata.

    :param curve: The curve, defaults to None
    :type curve: Curve, optional
    :param resolution: The resolution of the time series, defaults to None
    :type resolution: Resolution, optional
    :param instance: The instance, defaults to None
    :type instance: Instance, optional
    :param data: A list of periods (Period or CapacityPeriod)
    :type data: list[]
    """

    def __init__(self, data=None, **kwargs):
        super().__init__(**kwargs)
        assert isinstance(self.resolution, Resolution), (
            "Periodseries.resolution is required"
        )
        self.data = data or []

    def __str__(self):
        items = []
        items.append(f"resolution={self.resolution}")
        items.append(f"curve=\"{self.curve}\"")
        if self.instance:
            items.append(f"instance={self.instance}")
        if self.has_data():
            items.append(f"begin=\"{self.begin().isoformat(sep=' ')}\"")
            items.append(f"end=\"{self.end().isoformat(sep=' ')}\"")
        else:
            items.append("EMPTY")
        return f"<Periodseries: {', '.join(items)}>"

    def __repr__(self):
        return str(self)

    def begin(self):
        if self.data:
            return self.data[0].begin
        else:
            raise ValueError("Periodseries has no values")

    def end(self):
        if self.data:
            return self.data[-1].end
        else:
            raise ValueError("Periodseries has no values")

    def to_timeseries(self, frequency=None):
        """
        Convert this period-based series to a regular time series.

        When periods overlap the same step in the resulting time series,
        a weighted average is calculated down to second-precision.

        :param frequency: The frequency of the resulting time series
        :type frequency: Frequency, required
        :return: A time series
        :rtype: Timeseries
        """
        # Verify parameters
        assert isinstance(frequency, Frequency), "Must be a frequency"
        # Prepare conversion
        resolution = Resolution(frequency, self.resolution.timezone)
        if not self.has_data():
            return Timeseries(
                curve=self.curve,
                resolution=resolution,
                instance=self.instance,
                data=[]
            )
        begin = resolution.floor(self.begin())
        end = self.end()
        iterator = _PeriodsToTimeseriesIterator(
            periods=self.data,
            resolution=resolution,
            begin=begin,
            end=end
        )
        # Convert
        data = [Value(dt, value) for dt, value in iterator]
        timeseries = Timeseries(
            curve=self.curve,
            resolution=resolution,
            instance=self.instance,
            data=data
        )
        timeseries.set_name(self._name)
        return timeseries

    def to_df(self, frequency=None, name=None, single_level_header=False):
        """
        Alias for :meth:`Periodseries.to_dataframe`.

        Convert this period-based to a ``pandas.DataFrame`` as a time series
        in the given frequency.

        Using :py:meth:`Periodseries.to_timeseries` to convert this
        period-based series to a regular time series first. When periods
        overlap the same step in the resulting time series, a weighted
        average is calculated down to second-precision.

        :param frequency: The frequency of the resulting ``pandas.DataFrame``\
            time series
        :type frequency: Frequency, required
        :param name: Set a name for the column in the ``pandas.DataFrame``,\
            defaults to ``value``
        :type name: str, optional
        :param single_level_header: Set to True to use single-level header \
            in the DataFrame, defaults to False
        :type single_level_header: boolean, optional
        :return: A DataFrame
        :rtype: pandas.DataFrame
        :raises ImportError: When pandas is not installed on the system
        """
        return self.to_dataframe(
            frequency=frequency,
            name=name,
            single_level_header=single_level_header
        )

    def to_dataframe(self, frequency=None, name=None,
                     single_level_header=False):
        """
        Convert this period-based to a ``pandas.DataFrame`` as a time series
        in the given frequency.

        Using :py:meth:`Periodseries.to_timeseries` to convert this
        period-based series to a regular time series first. When periods
        overlap the same step in the resulting time series, a weighted
        average is calculated down to second-precision.

        :param frequency: The frequency of the resulting ``pandas.DataFrame``\
            time series
        :type frequency: Frequency, required
        :param name: Set a name for the column in the ``pandas.DataFrame``,\
            defaults to ``value``
        :type name: str, optional
        :param single_level_header: Set to True to use single-level header \
            in the DataFrame, defaults to False
        :type single_level_header: boolean, optional
        :return: A DataFrame
        :rtype: pandas.DataFrame
        :raises ImportError: When pandas is not installed on the system
        """
        # Verify parameters
        assert isinstance(frequency, Frequency), "Must be a frequency"
        # Conversion
        timeseries = self.to_timeseries(frequency=frequency)
        df = timeseries.to_dataframe(
            name=name,
            single_level_header=single_level_header
        )
        return df

    def print(self, file=sys.stdout):
        """
        Utility method to print a period-based series to any file handle
        (defaults to stdout).
        """
        print(f"Periodseries:", file=file)
        if self.curve:
            print(f"  Curve: {repr(self.curve)}", file=file)
        if self.instance:
            print(f"  Instance: {self.instance}", file=file)
        print(f"  Resolution: {self.resolution}", file=file)
        print(f"", file=file)
        for d in self.data:
            d.print(file=file)


class _PeriodsToTimeseriesIterator:
    """
    A period-based series iterator used for conversions to Timeseries objects.
    """

    def __init__(self, periods=None, resolution=None, begin=None, end=None):
        self.periods = [p for p in periods if p.end > begin and p.begin < end]
        self.resolution = resolution
        self.begin = begin
        self.end = end
        # Iterator stuff
        self.d = None
        self.p = None

    def __iter__(self):
        # No periods available
        if not self.periods:
            return []
        # Get first period
        self.d = self.begin
        self.p = self.periods.pop(0)
        return self

    def __next__(self):
        # Get dates and current period
        p = self.p
        d0 = self.d
        d1 = self.d = self.resolution >> d0
        # We're done
        if d0 >= self.end:
            raise StopIteration
        else:
            return self._find_next_value(p, d0, d1)

    def _find_next_value(self, p, d0, d1):
        # No more periods
        if not p:
            return (d0, None)
        # Period covers the entire time interval
        if p.is_covered(d0, d1):
            return (d0, p.value)
        # We do not have any values for given date
        if p.is_interval_before(d0, d1):
            return (d0, None)
        # We are past current period
        if p.is_interval_after(d0, d1):
            p = self.p = self.periods.pop(0) if self.periods else None
            return self._find_next_value(p, d0, d1)
        # Overlapping, but not covering – find all periods covering interval
        overlapping = self._get_overlayed_periods(d0, d1)
        # Current period starts in the middle of the interval
        if not overlapping:
            return (d0, None)
        # More than one period – check if they are connected (no gaps)
        if self._is_covering_interval(p, overlapping, d0, d1):
            # No gaps – generate a mean value
            mean = self._calc_mean_periods([p] + overlapping, d0, d1)
            return (d0, mean)
        # There are gaps, so we do not have a value
        return (d0, None)

    def _get_overlayed_periods(self, begin, end):
        # Find other periods overlapping current interval
        periods = []
        for p in self.periods:
            if p.is_overlayed(begin, end):
                periods.append(p)
            else:
                break
        return periods

    def _is_covering_interval(self, current, periods, begin, end):
        # Check if the first period starts somewhere in the interval
        if current.begin > begin:
            return False
        # Check for gaps between periods
        previous = current
        for p in periods:
            if previous.end != p.begin:
                return False
            previous = p
        # Check if the last element stops somewhere in the interval
        if previous.end < end:
            return False
        # All covered
        return True

    def _calc_mean_periods(self, periods, begin, end):
        # Get value and duration in interval for each period
        available_weights = [
            (p.value, p.get_duration_seconds(begin, end))
            for p in periods
        ]
        # Get sum of weight
        sum_weights = 1.0 * sum(weight for avail, weight in available_weights)
        # Get the mean value
        return (
            sum(avail * weight for avail, weight in available_weights)
            / sum_weights
        )


# Period-based series list with helpers



class PeriodseriesList(list):
    """
    A list of Periodseries objects. Have methods for converting them to a
    :py:class:`TimeseriesList` or a `pandas.DataFrame`.

    :param iterable: Any iterable of `Periodseries`
    :type iterable: iterable
    """

    def __init__(self, iterable=()):
        # Initialize list
        super().__init__(iterable)
        # Asserts
        _validate_periodseries_list(iterable)

    def to_timeseries(self, frequency=None):
        """
        Convert all period-based series in this list to time series.

        When periods overlap the same step in the resulting time series,
        a weighted average is calculated down to second-precision.

        :param frequency: The frequency of the resulting time series
        :type frequency: Frequency, required
        :return: A list of time series
        :rtype: TimeseriesList
        """
        # Verify parameters
        assert isinstance(frequency, Frequency), "Must be a frequency"
        # Convert all period-based series to time series
        return TimeseriesList(
            periodseries.to_timeseries(frequency=frequency)
            for periodseries in self
        )

    def to_df(self, frequency=None, single_level_header=False):
        """
        Alias for :meth:`Timeseries.to_dataframe`.

        Convert this PeriodseriesList to a ``pandas.DataFrame`` where all time
        series are placed in its own column and are lined up with the date-time
        as index.

        :param frequency: The frequency of the resulting time series'
        :type frequency: Frequency, required
        :param single_level_header: Set to True to use single-level header \
            in the DataFrame, defaults to False
        :type single_level_header: boolean, optional
        :return: A DataFrame
        :rtype: pandas.DataFrame
        :raises ImportError: When pandas is not installed on the system
        """
        return self.to_dataframe(
            frequency=frequency,
            single_level_header=single_level_header
        )

    def to_dataframe(self, frequency=None, single_level_header=False):
        """
        Convert this PeriodseriesList to a ``pandas.DataFrame`` where all time
        series are placed in its own column and are lined up with the date-time
        as index.

        :param frequency: The frequency of the resulting time series'
        :type frequency: Frequency, required
        :param single_level_header: Set to True to use single-level header \
            in the DataFrame, defaults to False
        :type single_level_header: boolean, optional
        :return: A DataFrame
        :rtype: pandas.DataFrame
        :raises ImportError: When pandas is not installed on the system
        """
        # Verify parameters
        assert isinstance(frequency, Frequency), "Must be a frequency"
        # Convert to time series then to data frame
        timeseries_list = self.to_timeseries(frequency=frequency)
        return timeseries_list.to_dataframe(
            single_level_header=single_level_header
        )

    def append(self, periodseries):
        _validate_periodseries(periodseries)
        return super().append(periodseries)

    def extend(self, iterable):
        # Asserts
        _validate_periodseries_list(iterable)
         # Perform operation
        return super().extend(iterable)

    def insert(self, index, periodseries):
        # Asserts
        _validate_periodseries(periodseries)
         # Perform operation
        return super().insert(index, periodseries)

    def __add__(self, rhs):
        _validate_periodseries_list(rhs)
        return PeriodseriesList(list.__add__(self, rhs))

    def __iadd__(self, rhs):
        _validate_periodseries_list(rhs)
        return PeriodseriesList(list.__iadd__(self, rhs))

    def __setitem__(self, key, periodseries):
        _validate_periodseries(periodseries)
        return super().__setitem__(periodseries)

    def __mul__(self, rhs):
        raise NotImplementedError("PeriodseriesList does not support multiply")

    def __rmul__(self, rhs):
        raise NotImplementedError("PeriodseriesList does not support multiply")

    def __imul__(self, rhs):
        raise NotImplementedError("PeriodseriesList does not support multiply")

    def copy(self):
        return PeriodseriesList(self)

    def __getitem__(self, item):
        result = list.__getitem__(self, item)
        if isinstance(result, list):
            return PeriodseriesList(result)
        else:
            return result


def _validate_periodseries(periodseries):
    assert isinstance(periodseries, Periodseries), (
        f"Element is not a Periodseries. Expects all "
        f"elements to be Periodseries objects."
    )


def _validate_periodseries_list(periodseries_list):
    for index, periodseries in enumerate(periodseries_list):
        assert isinstance(periodseries, Periodseries), (
            f"Element {index} is not a Periodseries. Expects all "
            f"elements to be Periodseries objects."
        )




class InitializationError(Exception):
    """
    Initialization error for the Energy Quantified Python client.
    """

    def __init__(self, reason=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reason = reason or ""

    def __str__(self):
        return self.reason


from .timeseries import (
    Timeseries,
    TimeseriesList,
    ValueType,
    Value,
    ScenariosValue,
    MeanScenariosValue
)
from .periodseries import (
    Periodseries,
    PeriodseriesList,
    Period,
    CapacityPeriod
)
from .ohlc import Product, OHLC, OHLCList
from .srmc import SRMC, SRMCOptions


__all__ = [
    # Time series
    "Timeseries",
    "TimeseriesList",
    "ValueType",
    "Value",
    "ScenariosValue",
    "MeanScenariosValue",
    # Period series
    "Periodseries",
    "PeriodseriesList",
    "Period",
    "CapacityPeriod",
    # OHLC
    "Product",
    "OHLC",
    "OHLCList",
    # SRMC
    "SRMCOptions",
    "SRMC",
]


# pylint: disable=pylint-enums-no-annotated-value
# pylint: disable=protected-access
# pylint: disable=too-many-instance-attributes
import enum


class Allocation(enum.Enum):
    """
    Enumerator of border allocation types between two price areas.
    """

    #: Explicit exchange border
    EXPLICIT = ("E", "Explicit")
    #: Implicit exchange border
    IMPLICIT = ("I", "Implicit")
    #: Flow-based exchange border
    FLOW_BASED = ("F", "Flow-based")
    #: No commercial exchange available
    NO_COMMERCIAL_CAPACITY = ("N", "No commercial capacity")

    def __init__(self, tag, label):
        self.tag = tag
        self.label = label

    def __repr__(self):
        return self.tag

    def __str__(self):
        return self.tag

    def __lt__(self, other):
        return self.tag.__lt__(other.tag)

    def __gt__(self, other):
        return self.tag.__gt__(other.tag)


class Area:
    """
    A representation of a price area or country.

    Areas do have a reference to all exchange borders with available
    commercial capacity allocation properties. Access the borders via
    ``area.borders[]``. If you just want direct access neighbour
    listing, use ``area.exchange_neighbours[]``.

    Some areas, such as Nord Pool, consist of smaller areas. In these
    cases, you can access their children via ``area.children[]``. You can
    also find the parent area via ``area.parent``.

    All areas has a ``tag``, which you will find in the curve names, and a
    full ``name``.
    """

    __lookup_tags = {}
    __enum_ordering = []

    @classmethod
    def is_valid_tag(cls, tag):
        """
        Check whether an area tag exists or not.

        :param tag: An area tag
        :type tag: str
        :return: True if it exists, otherwise False
        :rtype: bool
        """
        return tag.lower() in cls.__lookup_tags

    @classmethod
    def by_tag(cls, tag):
        """
        Look up area by tag.

        :param tag: An area tag
        :type tag: str
        :return: The area for the given tag
        :rtype: Area
        """
        return cls.__lookup_tags[tag.lower()]

    @classmethod
    def by_tags(cls, *tags):
        """
        Return multiple areas by tags.

        :return: A list of areas by the provided tags
        :rtype: list[Area]
        """
        return [cls.__lookup_tags[t.lower()] for t in tags]

    @classmethod
    def all(cls):
        """
        Return a list of all areas.

        :return: A list of all areas
        :rtype: list[Area]
        """
        return list(cls.__enum_ordering)

    def __init__(
        self,
        tag,
        name,
        country=False,
        price_area=False,
        control_area=False,
        external=False,
    ):
        # Validations
        assert tag is not None, "Parameter tag is missing"
        assert name is not None, "Parameter name is missing"
        assert tag not in self.__lookup_tags, "Duplicate tag '%s'" % tag
        # --- Attributes ---
        #: The area tag (used in curve names)
        self.tag = tag
        #: The full name of the area
        self.name = name
        #: True when this area is a country, otherwise False
        self.country = country
        #: True when this area is a price area, otherwise False
        self.price_area = price_area
        #: True when this area is a control area/TSO area, otherwise False
        self.control_area = control_area
        #: True when this area is outside of the supported region (used
        #: for exchange areas)
        self.external = external
        # Lookups and ordering
        Area.__lookup_tags[tag.lower()] = self
        Area.__enum_ordering.append(self)
        # --- Set of exchange neighbours ---
        #: Set of neighbouring areas
        self.exchange_neighbours = set()
        #: Set of exchange borders with exchange allocation types
        self.borders = set()
        # --- Parent/children relationship ---
        #: The parent area (set if this is a sub-area of another area)
        self.parent = None
        #: Set of child areas (set if this area is split into other,
        #: smaller areas)
        self.children = set()
        # Convert tag to variable name ("-" becomes "_")
        self._variable = self.tag.replace("-", "_")
        setattr(Area, self.tag.replace("-", "_"), self)

    _ordering_nb = []
    _ordering_borders = []
    _ordering_children = []
    _variable = None

    def get_family(self):
        """
        Get all descendants in a flat list for this Area.

        :return: A list of all descendants/sub-areas for this area
        :rtype: list[Area]
        """
        the_list = []
        self._family_internal(the_list)
        return the_list

    def _family_internal(self, the_list):
        the_list.append(self)
        for c in self.children:
            c._family_internal(the_list)

    def __repr__(self):
        return "<Area: %s>" % self.tag

    def __str__(self):
        return "<Area: %s>" % self.tag

    def __lt__(self, other):
        return self.tag.__lt__(other.tag)

    def __gt__(self, other):
        return self.tag.__gt__(other.tag)

    def _add_exchange_neighbours(self, *neighbours):
        """
        Private method. Add a list of exchange neighbours to self and
        self to the neighbour.
        """
        Area._ordering_nb.append(self)
        # Add neighbour relationships
        for nb in neighbours:
            assert nb != self, "Cannot add self to neighbour list: %s" % self.tag
            self.exchange_neighbours.add(nb)
            nb.exchange_neighbours.add(self)
        return self

    def _add_borders(self, *borders):
        """
        Private method. Add borders and exchange neighbours.
        """
        Area._ordering_borders.append(self)
        # Add exchange neighbours
        self._add_exchange_neighbours(*tuple(area for (area, allocations) in borders))
        # Add borders
        for nb, allocs in borders:
            assert nb != self, "Cannot add self to border list: %s" % self.tag
            allocations = tuple(a for a in list(Allocation) if a.tag in allocs)
            assert (
                allocations and len(allocations) > 0
            ), "At least one allocation required for %s -> %s" % (self, nb)
            self.borders.add(Border(self, nb, allocations))
            nb.borders.add(Border(nb, self, allocations))
        return self

    def _add_children(self, *children):
        """
        Private method. Add children to this area.
        """
        Area._ordering_children.append(self)
        # Add children to relationships
        for child in children:
            assert not child.parent, "Cannot add %s twice to children list: %s" % (
                child.tag,
                self.tag,
            )
            assert child != self, "Cannot add self to children list: %s" % self.tag
            child.parent = self
            self.children.add(child)


class Border:
    """
    A one-way border between two price areas.

    The `source` area is where the power is exported from, and the
    `sink` area is where to power is imported to.

    Each border has a list of capacity allocations describing how the
    commercial capacities on the border are set.
    """

    def __init__(self, source, sink, allocations=None):
        #: The source area (the exporter)
        self.source = source
        #: The sink area (the importer)
        self.sink = sink
        #: A tuple of the exchange allocation types (flow-based, implicit, etc.)
        self.allocations = allocations

    def is_explicit(self):
        """
        Returns True if this border has explicit allocations.

        :return: True if this border has explicit allocations, otherwise False
        :rtype: bool
        """
        return Allocation.EXPLICIT in self.allocations

    def is_implicit(self):
        """
        Returns True if this border has implicit allocations.

        :return: True if this border has implicit allocations, otherwise False
        :rtype: bool
        """
        return Allocation.IMPLICIT in self.allocations

    def is_flow_based(self):
        """
        Returns True if this border is flow-based.

        :return: True if this border is flow-based, otherwise False
        :rtype: bool
        """
        return Allocation.FLOW_BASED in self.allocations

    def as_tuple(self):
        """
        Convert this border to a tuple of (source, sink, allocations)

        :return: This border as a tuple of (source, sink, allocations)
        :rtype: tuple
        """
        return (self.source, self.sink, self.allocations)

    def __hash__(self):
        return hash(self.as_tuple())

    def __eq__(self, other):
        return self.as_tuple() == other.as_tuple()

    def __ne__(self, other):
        return self.as_tuple() != other.as_tuple()

    def __repr__(self):
        return "<Border: %s – %s, allocations=%s>" % (
            self.source.tag,
            self.sink.tag,
            sorted(self.allocations),
        )

    def __str__(self):
        return "<Border: %s – %s, allocations=%s>" % (
            self.source.tag,
            self.sink.tag,
            sorted(self.allocations),
        )


## Unknown area

NOAREA = Area(tag="", name="")


## Nordics

NP = Area(tag="NP", name="Nord Pool", price_area=True)

NO = Area(tag="NO", name="Norway", country=True)
NO1 = Area(tag="NO1", name="NO1 – Oslo", price_area=True)
NO2 = Area(tag="NO2", name="NO2 – Kristiansand", price_area=True)
NO3 = Area(tag="NO3", name="NO3 – Trondheim", price_area=True)
NO4 = Area(tag="NO4", name="NO4 – Tromsø", price_area=True)
NO5 = Area(tag="NO5", name="NO5 – Bergen", price_area=True)
NO._add_children(NO1, NO2, NO3, NO4, NO5)

SE = Area(tag="SE", name="Sweden", country=True)
SE1 = Area(tag="SE1", name="SE1 – Luleå", price_area=True)
SE2 = Area(tag="SE2", name="SE2 – Sundsvall", price_area=True)
SE3 = Area(tag="SE3", name="SE3 – Stockholm", price_area=True)
SE4 = Area(tag="SE4", name="SE4 – Malmö", price_area=True)
SE._add_children(SE1, SE2, SE3, SE4)

DK = Area(tag="DK", name="Denmark", country=True)
DK1 = Area(tag="DK1", name="DK1 – Jylland", price_area=True)
DK2 = Area(tag="DK2", name="DK2 – København", price_area=True)
DK._add_children(DK1, DK2)

FI = Area(tag="FI", name="Finland", country=True, price_area=True)

NP._add_children(NO, SE, DK, FI)


## Baltics

EE = Area(tag="EE", name="Estonia", country=True, price_area=True)
LV = Area(tag="LV", name="Latvia", country=True, price_area=True)
LT = Area(tag="LT", name="Lithuania", country=True, price_area=True)


## Central Western Europe

DE = Area(tag="DE", name="Germany", country=True, price_area=True)
DE_50Hertz = Area(tag="DE-50Hertz", name="Germany – 50Hertz", control_area=True)
DE_Amprion = Area(tag="DE-Amprion", name="Germany – Amprion", control_area=True)
DE_TenneT = Area(tag="DE-TenneT", name="Germany – TenneT GER", control_area=True)
DE_TransnetBW = Area(
    tag="DE-TransnetBW", name="Germany – TransnetBW", control_area=True
)
DE._add_children(DE_50Hertz, DE_Amprion, DE_TenneT, DE_TransnetBW)

FR = Area(tag="FR", name="France", country=True, price_area=True)
FR_ARA = Area(tag="FR-ARA", name="France – Auvergne-Rhône-Alpes", control_area=True)
FR_BFC = Area(tag="FR-BFC", name="France – Bourgogne-Franche-Comté", control_area=True)
FR_BRE = Area(tag="FR-BRE", name="France – Bretagne", control_area=True)
FR_COR = Area(tag="FR-COR", name="France – Corse", control_area=True)
FR_CVL = Area(tag="FR-CVL", name="France – Centre-Val de Loire", control_area=True)
FR_GES = Area(tag="FR-GES", name="France – Grand-Est", control_area=True)
FR_HDF = Area(tag="FR-HDF", name="France – Hauts-de-France", control_area=True)
FR_IDF = Area(tag="FR-IDF", name="France – Île-de-France", control_area=True)
FR_NAQ = Area(tag="FR-NAQ", name="France – Nouvelle-Aquitaine", control_area=True)
FR_NOR = Area(tag="FR-NOR", name="France – Normandie", control_area=True)
FR_OCC = Area(tag="FR-OCC", name="France – Occitanie", control_area=True)
FR_PAC = Area(
    tag="FR-PAC", name="France – Provence-Alpes-Côte d'Azur", control_area=True
)
FR_PDL = Area(tag="FR-PDL", name="France – Pays de la Loire", control_area=True)
FR._add_children(
    FR_ARA,
    FR_BFC,
    FR_BRE,
    FR_CVL,
    FR_GES,
    FR_HDF,
    FR_IDF,
    FR_NAQ,
    FR_NOR,
    FR_OCC,
    FR_PAC,
    FR_PDL,  # Removed: FR_COR (Corsica)
)

NL = Area(tag="NL", name="Netherlands", country=True, price_area=True)
BE = Area(tag="BE", name="Belgium", country=True, price_area=True)
AT = Area(tag="AT", name="Austria", country=True, price_area=True)
CH = Area(tag="CH", name="Switzerland", country=True, price_area=True)


## United Kingdom/Ireland

# National Grid
GB = Area(tag="GB", name="Great Britain", country=True, price_area=True)

# Single Electricity Market (IE + NIE). IE = EirGrid. NIE = SONI.
SEM = Area(
    tag="SEM",
    name="Single Electricity Market – Ireland",
    country=False,
    price_area=True,
)
NIE = Area(
    tag="NIE", name="Northern Ireland", country=True, price_area=True, control_area=True
)
IE = Area(tag="IE", name="Ireland", country=True, price_area=True, control_area=True)
SEM._add_children(NIE, IE)


## Central Eastern Europe

PL = Area(tag="PL", name="Poland", country=True, price_area=True)
CZ = Area(tag="CZ", name="Czech Republic", country=True, price_area=True)
HU = Area(tag="HU", name="Hungary", country=True, price_area=True)
SK = Area(tag="SK", name="Slovakia", country=True, price_area=True)


## Iberian Peninsula

ES = Area(tag="ES", name="Spain", country=True, price_area=True)
PT = Area(tag="PT", name="Portugal", country=True, price_area=True)


## Southern Europe

IT = Area(tag="IT", name="Italy", country=True, price_area=True)
IT_NORD = Area(tag="IT-NORD", name="Italy – Northern", price_area=True)
IT_CNOR = Area(tag="IT-CNOR", name="Italy – Central-Northern", price_area=True)
IT_CSUD = Area(tag="IT-CSUD", name="Italy – Central-Southern", price_area=True)
IT_SUD = Area(tag="IT-SUD", name="Italy – Southern", price_area=True)
IT_CALA = Area(tag="IT-CALA", name="Italy – Calabria", price_area=True)
IT_SARD = Area(tag="IT-SARD", name="Italy – Sardegna", price_area=True)
IT_SICI = Area(tag="IT-SICI", name="Italy – Sicily", price_area=True)
IT._add_children(IT_NORD, IT_CNOR, IT_CSUD, IT_SUD, IT_CALA, IT_SARD, IT_SICI)


SI = Area(tag="SI", name="Slovenia", country=True, price_area=True)
RO = Area(tag="RO", name="Romania", country=True, price_area=True)
HR = Area(tag="HR", name="Croatia", country=True, price_area=True)
BA = Area(tag="BA", name="Bosnia and Herzegovina", country=True, price_area=True)
RS = Area(tag="RS", name="Serbia", country=True, price_area=True)
ME = Area(tag="ME", name="Montenegro", country=True, price_area=True)
BG = Area(tag="BG", name="Bulgaria", country=True, price_area=True)
AL = Area(tag="AL", name="Albania", country=True, price_area=True)
MK = Area(tag="MK", name="North Macedonia", country=True, price_area=True)
GR = Area(tag="GR", name="Greece", country=True, price_area=True)
XK = Area(tag="XK", name="Kosovo", country=True, control_area=True)


## South-Eastern Europe

TR = Area(tag="TR", name="Turkey", country=True, price_area=True)


## Other

RU = Area(tag="RU", name="Russia", country=True, external=True)
RU_KGD = Area(tag="RU-KGD", name="Kaliningrad (Russia)", country=True, external=True)
BY = Area(tag="BY", name="Belarus", country=True, external=True)
UA = Area(tag="UA", name="Ukraine", country=True, external=True)
MD = Area(tag="MD", name="Moldova", country=True, external=True)
MT = Area(tag="MT", name="Malta", country=True, external=True)


## Nordic borders

NO1._add_borders((NO2, "I"), (NO3, "I"), (NO5, "I"), (SE3, "I"))
NO2._add_borders((DE, "I"), (DK1, "I"), (GB, "E"), (NL, "I"), (NO1, "I"), (NO5, "I"))
NO3._add_borders((NO1, "I"), (NO4, "I"), (NO5, "I"), (SE2, "I"))
NO4._add_borders((FI, "N"), (NO3, "I"), (SE1, "I"), (SE2, "I"))
NO5._add_borders((NO1, "I"), (NO2, "I"), (NO3, "I"))

SE1._add_borders((FI, "I"), (NO4, "I"), (SE2, "I"))
SE2._add_borders((NO3, "I"), (NO4, "I"), (SE1, "I"), (SE3, "I"))
SE3._add_borders((DK1, "I"), (FI, "I"), (NO1, "I"), (SE2, "I"), (SE4, "I"))
SE4._add_borders((DE, "I"), (DK2, "I"), (LT, "I"), (PL, "I"), (SE3, "I"))

DK1._add_borders((DE, "I"), (DK2, "I"), (NL, "I"), (NO2, "I"), (SE3, "I"))
DK2._add_borders((DE, "I"), (DK1, "I"), (SE4, "I"))

FI._add_borders((EE, "I"), (NO4, "N"), (RU, "E"), (SE1, "I"), (SE3, "I"))


## Baltic borders

EE._add_borders((FI, "I"), (LV, "I"), (RU, "N"))
LV._add_borders((EE, "I"), (LT, "I"), (RU, "E"))
LT._add_borders((BY, "N"), (LV, "I"), (PL, "I"), (RU_KGD, "E"), (SE4, "I"))


## Central Western Europe

DE._add_borders(
    (AT, "F"),
    (BE, "F"),
    (CH, "E"),
    (CZ, "EI"),
    (DK1, "I"),
    (DK2, "I"),
    (FR, "F"),
    (NL, "F"),
    (NO2, "I"),
    (PL, "EI"),
    (SE4, "I"),
)
AT._add_borders((CH, "E"), (CZ, "EI"), (DE, "F"), (HU, "EI"), (IT_NORD, "EI"), (SI, "EI"))
FR._add_borders(
    (BE, "F"), (CH, "E"), (DE, "F"), (ES, "EI"), (GB, "E"), (IT_NORD, "EI")
)
FR_COR._add_borders((IT_SARD, "I"))
NL._add_borders((BE, "F"), (DE, "F"), (DK1, "I"), (GB, "E"), (NO2, "I"))
BE._add_borders((DE, "F"), (FR, "F"), (GB, "E"), (NL, "F"))
CH._add_borders((AT, "E"), (DE, "E"), (FR, "E"), (IT_NORD, "E"))


## United Kingdom/Ireland

GB._add_borders((BE, "E"), (FR, "E"), (NL, "E"), (NIE, "I"), (IE, "I"), (NO2, "E"))
NIE._add_borders((IE, "I"), (GB, "I"))
IE._add_borders((NIE, "I"), (GB, "I"))


## Central Eastern Europe

PL._add_borders((CZ, "EI"), (DE, "EI"), (LT, "I"), (SE4, "I"), (SK, "EI"), (UA, "E"))
CZ._add_borders((AT, "EI"), (DE, "EI"), (PL, "EI"), (SK, "EI"))
HU._add_borders((AT, "EI"), (HR, "E"), (RO, "EI"), (RS, "E"), (SK, "EI"), (UA, "E"))
SK._add_borders((CZ, "EI"), (HU, "EI"), (PL, "EI"), (UA, "E"))


## Iberian Peninsula

ES._add_borders((FR, "EI"), (PT, "I"))
PT._add_borders((ES, "I"))


## Southern Europe

IT_NORD._add_borders((AT, "EI"), (CH, "E"), (FR, "EI"), (IT_CNOR, "I"), (SI, "EI"))
IT_CNOR._add_borders((IT_CSUD, "I"), (IT_NORD, "I"), (IT_SARD, "I"))
IT_CSUD._add_borders((IT_CNOR, "I"), (IT_SARD, "I"), (IT_SUD, "I"), (ME, "E"))
IT_SUD._add_borders((GR, "EI"), (IT_CSUD, "I"), (IT_CALA, "I"))
IT_CALA._add_borders((IT_SUD, "I"), (IT_SICI, "I"))
IT_SICI._add_borders((IT_CALA, "I"), (MT, "I"))
IT_SARD._add_borders((FR_COR, "I"), (IT_CNOR, "I"), (IT_CSUD, "I"))


SI._add_borders((AT, "EI"), (HR, "EI"), (IT_NORD, "EI"))
RO._add_borders((BG, "E"), (HU, "EI"), (RS, "E"), (UA, "E"))
HR._add_borders((BA, "E"), (HU, "E"), (RS, "E"), (SI, "EI"))
BA._add_borders((HR, "E"), (ME, "E"), (RS, "E"))
RS._add_borders(
    (BA, "E"),
    (BG, "E"),
    (HR, "E"),
    (HU, "E"),
    (ME, "E"),
    (MK, "E"),
    (RO, "E"),
)
ME._add_borders((AL, "E"), (BA, "E"), (IT_CSUD, "E"), (RS, "E"), (XK, "E"))
BG._add_borders((GR, "EI"), (MK, "E"), (RO, "E"), (RS, "E"), (TR, "E"))
AL._add_borders((GR, "E"), (ME, "E"), (XK, "E"))
MK._add_borders((BG, "E"), (GR, "E"), (RS, "E"), (XK, "E"))
GR._add_borders((AL, "E"), (BG, "EI"), (IT_SUD, "EI"), (MK, "E"), (TR, "E"))
XK._add_borders((AL, "E"), (ME, "E"), (MK, "E"))


## South-Eastern Europe

TR._add_borders((BG, "E"), (GR, "E"))


## Others

RU._add_borders((EE, "N"), (FI, "E"), (LV, "E"))
RU_KGD._add_borders((LT, "E"))
BY._add_borders((LT, "N"))
UA._add_borders((HU, "E"), (PL, "E"), (RO, "E"), (SK, "E"))
MT._add_borders((IT_SICI, "I"))





def dict_to_str(some_dict, prefix=""):
    """
    Converts a dict to a deterministic string for dicts that have the same
    keys and values (the string will be the same regardless of the original
    ordering of the keys).
    """
    keys_sorted = sorted(some_dict.keys())
    parts = [prefix]
    for k in keys_sorted:
        parts.append("|")
        parts.append(str(k))
        parts.append("=")
        parts.append(str(some_dict[k]))
    return "".join(parts)